#!/usr/bin/env python

import logging
l = logging.getLogger("claripy.frontends.light_frontend")

from ..frontend import Frontend

class LightFrontend(Frontend):
    def __init__(self, solver_backend):
        Frontend.__init__(self, solver_backend)
        self.constraints = [ ]
        self.variables = set()
        self._finalized = False

    #
    # Storable support
    #

    def _ana_getstate(self):
        if not self._simplified: self.simplify()
        self.finalize()
        return self.constraints, self.variables, Frontend._ana_getstate(self)

    def _ana_setstate(self, s):
        self.constraints, self.variables, base_state = s
        Frontend._ana_setstate(self, base_state)
        self._finalized = True

    #
    # Constraint management
    #

    def independent_constraints(self):
        return self._split_constraints(self.constraints)

    #
    # Light functionality
    #

    def _add_constraints(self, constraints, invalidate_cache=True):
        self.constraints += constraints
        for c in constraints:
            self.variables.update(c.variables)
        return constraints


    def _simplify(self):
        if len(self.constraints) == 0:
            return

        self.constraints = [ simplify(And(*self.constraints)) ]

        # generate UUIDs for every constraint
        for c in self.constraints:
            if isinstance(c, Base): c.make_uuid()

        self._simplified = True
        return self.constraints

    def _solve(self, extra_constraints=()):
        return SatResult(approximation=True)

    def _satisfiable(self, extra_constraints=()):
        return self.solve(extra_constraints=extra_constraints).sat

    def _eval(self, e, n, extra_constraints=()):
        if len(extra_constraints) == 0:
            for b in _eager_backends + [ self._solver_backend ]:
                try: return b.eval(e, n, result=self.result)
                except BackendError: pass

        raise ClaripyFrontendError("Light solver can't handle this eval().")

    def _max(self, e, extra_constraints=()):
        if len(extra_constraints) == 0:
            for b in _eager_backends + [ self._solver_backend ]:
                try: return b.max(e, result=self.result)
                except BackendError: pass

        raise ClaripyFrontendError("Light solver can't handle this max().")

    def _min(self, e, extra_constraints=()):
        extra_constraints = self._constraint_filter(extra_constraints)

        if len(extra_constraints) == 0:
            for b in _eager_backends + [ self._solver_backend ]:
                try: return b.min(e, result=self.result)
                except BackendError: pass

        two = self.eval(e, 2, extra_constraints=extra_constraints)
        if len(two) == 0: raise UnsatError("unsat during min()")
        elif len(two) == 1: return two[0]

        raise ClaripyFrontendError("Light solver can't handle this min().")

    def _solution(self, e, v, extra_constraints=()):
        if len(extra_constraints) == 0:
            for b in _eager_backends + [ self._solver_backend ]:
                try: return b.solution(e, v, result=self.result)
                except BackendError: pass

        raise ClaripyFrontendError("Light solver can't handle this solution().")

    #
    # Serialization and such.
    #

    def downsize(self):
        Frontend.downsize(self)

    #
    # Merging and splitting
    #

    def finalize(self):
        self._finalized = True

    def branch(self):
        s = Frontend.branch(self)
        s.constraints = list(self.constraints)
        s.variables = set(self.variables)
        self.finalize()
        s.finalize()
        return s

    def merge(self, others, merge_flag, merge_values):
        merged = self.__class__(self._solver_backend)
        merged._simplified = False
        options = [ ]

        for s, v in zip([self]+others, merge_values):
            options.append(And(*([ merge_flag == v ] + s.constraints)))
        merged.add([Or(*options)])

        return self._solver_backend is backend_z3, merged

    def combine(self, others):
        combined = self.__class__(self._solver_backend)
        combined._simplified = False

        combined.add(self.constraints) #pylint:disable=E1101
        for o in others:
            combined.add(o.constraints)
        return combined

    def split(self):
        results = [ ]
        l.debug("Splitting!")
        for variables,c_list in self.independent_constraints():
            l.debug("... got %d constraints with %d variables", len(c_list), len(variables))

            s = self.__class__(self._solver_backend)
            s._simplified = False
            s.add(c_list)
            results.append(s)
        return results

from ..result import SatResult
from ..errors import UnsatError, BackendError, ClaripyFrontendError
from .. import _eager_backends, backend_z3
from ..ast.base import Base, simplify
from ..ast.bool import And, Or



import logging
l = logging.getLogger("claripy.frontends.composite_frontend")

import itertools
symbolic_count = itertools.count()

from ..frontend import Frontend
from .full_frontend import FullFrontend

class CompositeFrontend(Frontend):
    def __init__(self, solver_backend, timeout=None, solver_class=FullFrontend):
        Frontend.__init__(self, solver_backend)
        self.timeout = timeout
        self._solvers = { }
        self._solver_class = solver_class

        self._solvers['CONCRETE'] = self._merged_solver_for({'CONCRETE'})

    #
    # Serialization stuff
    #

    def _ana_getstate(self):
        return self._solvers, self._solver_class, Frontend._ana_getstate(self)

    def _ana_setstate(self, s):
        self._solvers, self._solver_class, base_state = s
        Frontend._ana_setstate(base_state)

    #
    # Frontend management
    #

    @property
    def _solver_list(self):
        seen_solvers = set()
        solver_list = [ ]
        for s in self._solvers.itervalues():
            if id(s) in seen_solvers: continue
            seen_solvers.add(id(s))
            solver_list.append(s)
        return solver_list

    @property
    def variables(self):
        return set(self._solvers.keys()) - { 'CONCRETE' }

    @property
    def constraints(self):
        return sum([ s.constraints for s in self._solver_list ], [ ])

    def _solvers_for_variables(self, names):
        seen_solvers = set()
        existing_solvers = [ ]
        for n in names:
            if n not in self._solvers: continue
            s = self._solvers[n]

            if id(s) in seen_solvers: continue
            seen_solvers.add(id(s))
            existing_solvers.append(s)
        return existing_solvers

    def _merged_solver_for(self, names):
        l.debug("composite_solver._merged_solver_for() running with %d names", len(names))
        solvers = self._solvers_for_variables(names)
        if len(solvers) == 0:
            l.debug("... creating new solver")
            return self._solver_class(self._solver_backend, timeout=self.timeout)
        elif len(solvers) == 1:
            l.debug("... got one solver")
            return solvers[0]
        else:
            l.debug(".... combining %d solvers", len(solvers))
            return solvers[0].combine(solvers[1:])

    def _shared_solvers(self, others):
        '''
        Returns a sequence of the solvers that self and others share.
        '''

        solvers_by_id = { s.uuid: s for s in self._solver_list }
        common_solvers = set(solvers_by_id.keys())
        other_sets = [ { s.uuid for s in cs._solver_list } for cs in others ]
        for o in other_sets: common_solvers &= o

        return [ solvers_by_id[s] for s in common_solvers ]

    def _variable_sets(self):
        return { s.variables for s in self._solver_list }

    def _shared_varsets(self, others):
        common_varsets = self._variable_sets()
        for o in others: common_varsets &= o.all_varsets()
        return common_varsets

    #
    # Constraints
    #

    def _add_dependent_constraints(self, names, constraints):
        l.debug("Adding %d constraints to %d names", len(constraints), len(names))
        s = self._merged_solver_for(names)
        s.add(constraints)
        for v in s.variables | names:
            self._solvers[v] = s

    def _add_constraints(self, constraints, invalidate_cache=True):
        if not invalidate_cache:
            l.warning("ignoring non-invalidating constraints")
            return

        split = self._split_constraints(constraints)

        l.debug("%s, solvers before: %d", self, len(self._solvers))
        for names,set_constraints in split:
            self._add_dependent_constraints(names, set_constraints)
        l.debug("... solvers after add: %d", len(self._solver_list))

    #
    # Solving
    #

    def _solve(self, extra_constraints=()):
        l.debug("%r checking satisfiability...", self)

        if len(extra_constraints) != 0:
            extra_vars = frozenset.union(*(a.variables for a in extra_constraints))
            solvers = [ self._merged_solver_for(extra_vars) ]
            for s in self._solver_list:
                if len(s.variables | solvers[0].variables) == 0:
                    solvers.append(s)
        else:
            solvers = self._solver_list

        model = { }
        satness = True

        for s in solvers:
            if not s.satisfiable(extra_constraints=extra_constraints if s is solvers[0] else ()):
                l.debug("... %r: False", s)
                satness = False
                break

            l.debug("... %r: True", s)
            model.update(s.result.model)

        l.debug("... ok!")
        return Result(satness, model=model)

    def _all_variables(self, e, extra_constraints=()): #pylint:disable=no-self-use
        all_vars = e.variables
        if len(extra_constraints) != 0:
            all_vars |= frozenset.union(*(a.variables for a in extra_constraints))
        if len(all_vars) == 0:
            all_vars = { 'CONCRETE' }
        return all_vars

    def _eval(self, e, n, extra_constraints=()):
        return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).eval(e, n, extra_constraints=extra_constraints)

    def _max(self, e, extra_constraints=()):
        return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).max(e, extra_constraints=extra_constraints)

    def _min(self, e, extra_constraints=()):
        return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).min(e, extra_constraints=extra_constraints)

    def _solution(self, e, n, extra_constraints=()):
        return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).solution(e, n, extra_constraints=extra_constraints)

    def _simplify(self):
        l.debug("Simplifying %r with %d solvers", self, len(self._solver_list))
        for s in self._solver_list:
            if s._simplified:
                continue

            l.debug("... simplifying child solver %r", s)
            s.simplify()
            l.debug("... splitting child solver %r", s)
            split = s.split()

            l.debug("... split solver %r into %d parts", s, len(split))
            l.debug("... variable counts: %s", [ len(ss.variables) for ss in split ])
            if len(split) > 1:
                for s in split:
                    for v in s.variables:
                        self._solvers[v] = s
        l.debug("... after-split, %r has %d solvers", self, len(self._solver_list))

    #
    # Merging and splitting
    #

    def finalize(self):
        l.error("CompositeFrontend.finalize is incomplete. This represents a big issue.")
        for s in self._solver_list:
            s.finalize()

    def branch(self):
        c = Frontend.branch(self)
        c.timeout = self.timeout

        for s in self._solver_list:
            c_s = s.branch()
            for v in c_s.variables:
                c._solvers[v] = c_s #pylint:disable=no-member

        if self.result is not None:
            c.result = self.result.branch()

        return c

    def merge(self, others, merge_flag, merge_values):
        l.debug("Merging %s with %d other solvers.", self, len(others))
        merged = CompositeFrontend(self._solver_backend)
        common_solvers = self._shared_solvers(others)
        common_ids = { s.uuid for s in common_solvers }
        l.debug("... %s common solvers", len(common_solvers))

        for s in common_solvers:
            for v in s.variables:
                merged._solvers[v] = s.branch()

        noncommon_solvers = [ [ s for s in cs._solver_list if s.uuid not in common_ids ] for cs in [self]+others ]

        l.debug("... merging noncommon solvers")
        combined_noncommons = [ ]
        for ns in noncommon_solvers:
            l.debug("... %d", len(ns))
            if len(ns) == 0:
                s = self._solver_class(self._solver_backend)
                s.timeout = self.timeout
                s.add(True)
                combined_noncommons.append(s)
            elif len(ns) == 1:
                combined_noncommons.append(ns[0])
            else:
                combined_noncommons.append(ns[0].combine(ns[1:]))

        _, merged_noncommon = combined_noncommons[0].merge(combined_noncommons[1:], merge_flag, merge_values)
        for v in merged_noncommon.variables:
            merged._solvers[v] = merged_noncommon

        return True, merged

    def combine(self, others):
        combined = self.__class__(self._solver_backend)
        combined._simplified = False
        combined.add(self.constraints)
        for o in others:
            combined.add(o.constraints)
        return combined

    def split(self):
        return [ s.branch() for s in self._solver_list if len(s.variables) > 0 ]

from ..result import Result



import sys
import logging
l = logging.getLogger("claripy.backends.backend_z3")

if sys.platform == 'darwin':
    z3_library_file = "libz3.dylib"
elif sys.platform == 'win32':
    z3_library_file = "libz3.dll"
else:
    z3_library_file = "libz3.so"


solve_count = 0
cache_count = 0

from decimal import Decimal
import weakref

# import and set up Z3
import os
import z3

from ..errors import ClaripyZ3Error

_z3_paths = [ ]

if "Z3PATH" in os.environ:
    _z3_paths.append(os.environ["Z3PATH"])
if "VIRTUAL_ENV" in os.environ:
    virtual_env = os.environ["VIRTUAL_ENV"]
    _z3_paths.append(os.path.join(os.environ["VIRTUAL_ENV"], "lib"))
_z3_paths.extend(sys.path)
_z3_paths.append("/usr/lib")
_z3_paths.append("/usr/local/lib")
_z3_paths.append("/opt/python/lib")

for z3_path in _z3_paths:
    if not '.so' in z3_path and not '.dll' in z3_path:
        z3_path = os.path.join(z3_path, z3_library_file)
    if os.path.exists(z3_path):
        z3.init(z3_path)
        break
else:
    raise ClaripyZ3Error("Unable to find %s", z3_library_file)

supports_fp = hasattr(z3, 'fpEQ')

from ..backend import Backend

#pylint:disable=unidiomatic-typecheck

#import threading
#import functools
#z3_lock = threading.RLock()
#def synchronized(f):
#      @functools.wraps(f)
#      def synced(self, *args, **kwargs):
#          if not (self._background_solve or (self._background_solve is None and self._claripy.parallel)):
#              return f(self, *args, **kwargs)
#
#          try:
#              #while not z3_lock.acquire(blocking=False): print "ACQUIRING...",__import__('time').sleep(1)
#              z3_lock.acquire()
#              return f(self, *args, **kwargs)
#          finally:
#              z3_lock.release()
#      return synced

def condom(f):
    def z3_condom(*args, **kwargs):
        '''
        The Z3 condom intersects Z3Exceptions and throws a ClaripyZ3Error instead.
        '''
        try:
            args = tuple((int(a) if type(a) is long and a < sys.maxint else a) for a in args)
            return f(*args, **kwargs)
        except z3.Z3Exception as ze:
            raise ClaripyZ3Error("Z3Exception: %s" % ze)
    return z3_condom

class BackendZ3(Backend):
    _split_on = { 'And', 'Or' }

    def __init__(self):
        Backend.__init__(self, solver_required=True)
        self._enable_simplification_cache = False

        # and the operations
        all_ops = backend_fp_operations | backend_operations if supports_fp else backend_operations
        for o in all_ops - {'Reverse', 'fpToSBV', 'fpToUBV', 'SLT', 'SLE', 'SGT', 'SGE'}:
            self._op_raw[o] = getattr(z3, o)

        self._op_raw['SLT'] = self.SLT
        self._op_raw['SLE'] = self.SLE
        self._op_raw['SGT'] = self.SGT
        self._op_raw['SGE'] = self.SGE
        self._op_raw['Reverse'] = self.reverse
        self._op_raw['Identical'] = self.identical
        self._op_raw['I'] = lambda thing: thing
        self._op_raw['fpToSBV'] = self.fpToSBV
        self._op_raw['fpToUBV'] = self.fpToUBV

    @property
    def _ast_cache(self):
        try:
            return self._tls.ast_cache
        except AttributeError:
            self._tls.ast_cache = weakref.WeakValueDictionary()
            return self._tls.ast_cache

    @property
    def _var_cache(self):
        try:
            return self._tls.var_cache
        except AttributeError:
            self._tls.var_cache = weakref.WeakValueDictionary()
            return self._tls.var_cache

    @property
    def _sym_cache(self):
        try:
            return self._tls.sym_cache
        except AttributeError:
            self._tls.sym_cache = weakref.WeakValueDictionary()
            return self._tls.sym_cache

    @property
    def _simplification_cache_key(self):
        try:
            return self._tls.simplification_cache_key
        except AttributeError:
            self._tls.simplification_cache_key = weakref.WeakValueDictionary()
            return self._tls.simplification_cache_key

    @property
    def _simplification_cache_val(self):
        try:
            return self._tls.simplification_cache_val
        except AttributeError:
            self._tls.simplification_cache_val = weakref.WeakValueDictionary()
            return self._tls.simplification_cache_val

    def downsize(self):
        Backend.downsize(self)

        self._ast_cache.clear()
        self._var_cache.clear()
        self._sym_cache.clear()
        self._simplification_cache_key.clear()
        self._simplification_cache_val.clear()

    @condom
    def _size(self, e, result=None):
        if not isinstance(e, z3.BitVecRef) and not isinstance(e, z3.BitVecNumRef):
            l.debug("Unable to determine length of value of type %s", e.__class__)
            raise BackendError("Unable to determine length of value of type %s" % e.__class__)
        return e.size()

    def _name(self, e, result=None): #pylint:disable=unused-argument
        l.warning("BackendZ3.name() called. This is weird.")
        raise BackendError("name is not implemented yet")

    @staticmethod
    @condom
    def reverse(a):
        if a.size() == 8:
            return a
        elif a.size() % 8 != 0:
            raise ClaripyOperationError("can't reverse non-byte sized bitvectors")
        else:
            return z3.Concat(*[z3.Extract(i+7, i, a) for i in range(0, a.size(), 8)])

    @staticmethod
    @condom
    def SLT(a, b):
        return a < b

    @staticmethod
    @condom
    def SLE(a, b):
        return a <= b

    @staticmethod
    @condom
    def SGT(a, b):
        return a > b

    @staticmethod
    @condom
    def SGE(a, b):
        return a >= b

    @staticmethod
    @condom
    def fpToSBV(rm, fp, bv_len):
        return z3.fpToSBV(rm, fp, z3.BitVecSort(bv_len))

    @staticmethod
    @condom
    def fpToUBV(rm, fp, bv_len):
        return z3.fpToUBV(rm, fp, z3.BitVecSort(bv_len))

    def _identical(self, a, b, result=None):
        return a.eq(b)

    @condom
    def _convert(self, obj, result=None):
        if type(obj) is NativeBVV:
            return z3.BitVecVal(obj.value, obj.bits)
        elif isinstance(obj, FSort):
            return z3.FPSort(obj.exp, obj.mantissa)
        elif isinstance(obj, RM):
            if obj == RM_RNE:
                return z3.RNE()
            elif obj == RM_RNA:
                return z3.RNA()
            elif obj == RM_RTP:
                return z3.RTP()
            elif obj == RM_RTN:
                return z3.RTN()
            elif obj == RM_RTZ:
                return z3.RTZ()
            else:
                raise BackendError("unrecognized rounding mode")
        elif isinstance(obj, NativeFPV):
            val = str(obj.value)
            sort = self._convert(obj.sort)
            if val == 'inf':
                return z3.fpPlusInfinity(sort)
            elif val == '-inf':
                return z3.fpMinusInfinity(sort)
            elif val == '0.0':
                return z3.fpPlusZero(sort)
            elif val == '-0.0':
                return z3.fpMinusZero(sort)
            elif val == 'nan':
                return z3.fpNaN(sort)
            else:
                better_val = str(Decimal(obj.value))
                return z3.FPVal(better_val, sort)
        elif obj is True:
            return z3.BoolVal(True)
        elif obj is False:
            return z3.BoolVal(False)
        elif type(obj) in (int, long, float, str):
            return obj
        elif hasattr(obj, '__module__') and obj.__module__ == 'z3':
            return obj
        else:
            l.debug("BackendZ3 encountered unexpected type %s", type(obj))
            raise BackendError("unexpected type %s encountered in BackendZ3" % type(obj))

    @condom
    def call(self, *args, **kwargs):
        return Backend.call(self, *args, **kwargs)

    @condom
    def _abstract(self, z):
        #return self._abstract(z, split_on=split_on)[0]
        return self._abstract_internal(z.ctx.ctx, z.ast)

    @staticmethod
    def _z3_ast_hash(ctx, ast):
        """
        This is a better hashing function for z3 Ast objects. Z3_get_ast_hash() creates too many hash collisions.

        :param ctx: z3 Context
        :param ast: z3 Ast object
        :return: An integer - the hash
        """

        z3_hash = z3.Z3_get_ast_hash(ctx, ast)
        z3_ast_ref = ast.value # this seems to be the memory address
        z3_sort = z3.Z3_get_sort(ctx, ast).value
        return hash("%d_%d_%d" % (z3_hash, z3_sort, z3_ast_ref))

    def _abstract_internal(self, ctx, ast, split_on=None):
        h = self._z3_ast_hash(ctx, ast)
        try:
            return self._ast_cache[h]
        except KeyError:
            pass

        decl = z3.Z3_get_app_decl(ctx, ast)
        decl_num = z3.Z3_get_decl_kind(ctx, decl)
        z3_sort = z3.Z3_get_sort(ctx, ast)

        if decl_num not in z3_op_nums:
            raise ClaripyError("unknown decl kind %d" % decl_num)
        if z3_op_nums[decl_num] not in op_map:
            raise ClaripyError("unknown decl op %s" % z3_op_nums[decl_num])
        op_name = op_map[z3_op_nums[decl_num]]

        num_args = z3.Z3_get_app_num_args(ctx, ast)
        split_on = self._split_on if split_on is None else split_on
        new_split_on = split_on if op_name in split_on else set()
        children = [ self._abstract_internal(ctx, z3.Z3_get_app_arg(ctx, ast, i), new_split_on) for i in range(num_args) ]

        append_children = True

        if op_name == 'True':
            return BoolI(True)
        elif op_name == 'False':
            return BoolI(False)
        elif op_name.startswith('RM_'):
            return RM.from_name(op_name)
        elif op_name == 'BitVecVal':
            bv_num = long(z3.Z3_get_numeral_string(ctx, ast))
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            return BVI(NativeBVV(bv_num, bv_size), length=bv_size)
        elif op_name == 'FPVal':
            # this is really imprecise
            fp_mantissa = float(z3.Z3_fpa_get_numeral_significand_string(ctx, ast))
            fp_exp = long(z3.Z3_fpa_get_numeral_exponent_string(ctx, ast))
            value = fp_mantissa * (2 ** fp_exp)

            ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(ebits, sbits)

            return FPI(NativeFPV(value, sort))
        elif op_name in ('MinusZero', 'MinusInf', 'PlusZero', 'PlusInf', 'NaN'):
            ebits = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            sbits = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(ebits, sbits)

            if op_name == 'MinusZero':
                return FPI(NativeFPV(-0.0, sort))
            elif op_name == 'MinusInf':
                return FPI(NativeFPV(float('-inf'), sort))
            elif op_name == 'PlusZero':
                return FPI(NativeFPV(0.0, sort))
            elif op_name == 'PlusInf':
                return FPI(NativeFPV(float('inf'), sort))
            elif op_name == 'NaN':
                return FPI(NativeFPV(float('nan'), sort))
        elif op_name == 'UNINTERPRETED' and num_args == 0: # this *might* be a BitVec ;-)
            bv_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)

            #if bv_name.count('_') < 2:
            #      import ipdb; ipdb.set_trace()
            return BV("BitVec", (bv_name, bv_size), length=bv_size, variables={ bv_name }, symbolic=True)
        elif op_name == 'UNINTERPRETED':
            mystery_name = z3.Z3_get_symbol_string(ctx, z3.Z3_get_decl_name(ctx, decl))
            args = [ ]
            if mystery_name == 'bvsmod_i':
                op_name = '__mod__'
                decl_num = z3.Z3_OP_BSMOD
            elif mystery_name == 'bvsdiv_i':
                op_name = '__div__'
                decl_num = z3.Z3_OP_BSDIV
            else:
                l.error("Mystery operation %s in BackendZ3._abstract_internal. Please report this.", mystery_name)
        elif op_name == 'Extract':
            hi = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            lo = z3.Z3_get_decl_int_parameter(ctx, decl, 1)
            args = [ hi, lo ]
        elif op_name in ('SignExt', 'ZeroExt'):
            num = z3.Z3_get_decl_int_parameter(ctx, decl, 0)
            args = [ num ]
        elif op_name in ('fpToFP', 'fpToFPSigned'):
            exp = z3.Z3_fpa_get_ebits(ctx, z3_sort)
            mantissa = z3.Z3_fpa_get_sbits(ctx, z3_sort)
            sort = FSort.from_params(exp, mantissa)
            args = children + [sort]
            append_children = False
        elif op_name in ('fpToSBV', 'fpToUBV'):
            # uuuuuugggggghhhhhh
            bv_size = z3.Z3_get_bv_sort_size(ctx, z3_sort)
            args = children + [bv_size]
            append_children = False
        else:
            args = [ ]

        if append_children:
            args.extend(children)

        # fix up many-arg __add__
        if op_name in bin_ops and len(args) > 2:
            many_args = args #pylint:disable=unused-variable
            last = args[-1]
            rest = args[:-1]

            a = args[0].make_like(op_name, rest[:2])
            for b in rest[2:]:
                a = args[0].make_like(op_name, [a,b])
            args = [ a, last ]

        # hmm.... honestly not sure what to do here
        result_ty = op_type_map[z3_op_nums[decl_num]]
        ty = type(args[-1])

        if op_name == 'If':
            # If is polymorphic and thus must be handled specially
            ty = type(args[1])

            a = ty('If', tuple(args), length=args[1].length)
        else:
            if hasattr(ty, op_name) or hasattr(_all_operations, op_name):
                op = getattr(ty if hasattr(ty, op_name) else _all_operations, op_name)
                if op.calc_length is not None:
                    length = op.calc_length(*args)
                    a = result_ty(op_name, tuple(args), length=length)
                else:
                    a = result_ty(op_name, tuple(args))
            else:
                a = result_ty(op_name, tuple(args))

        self._ast_cache[h] = a
        return a

    def solver(self, timeout=None):
        s = z3.Solver()
        if timeout is not None:
            if 'soft_timeout' in str(s.param_descrs()):
                s.set('soft_timeout', timeout)
                s.set('solver2_timeout', timeout)
            else:
                s.set('timeout', timeout)
        return s

    def _add(self, s, c):
        s.add(*c)

    @condom
    def _check(self, s, extra_constraints=()):
        return self._check_and_model(s, extra_constraints=extra_constraints)[0]

    @condom
    def _check_and_model(self, s, extra_constraints=()): #pylint:disable=no-self-use
        global solve_count
        solve_count += 1
        if len(extra_constraints) > 0:
            s.push()
            s.add(*extra_constraints)

        l.debug("Doing a check!")
        #print "CHECKING"
        satness = s.check() == z3.sat
        if satness:
            model = s.model()
        else:
            model = None
        #print "CHECKED"

        if len(extra_constraints) > 0:
            s.pop()
        return satness, model

    @condom
    def _results(self, s, extra_constraints=(), generic_model=True):
        satness, z3_model = self._check_and_model(s, extra_constraints=extra_constraints)
        model = { }

        if satness:
            l.debug("sat!")
            if generic_model:
                for m_f in z3_model:
                    n = m_f.name()
                    m = m_f()
                    model[n] = model_object(z3_model.eval(m))
        else:
            l.debug("unsat!")

        return Result(satness, model, backend_model=z3_model)

    @condom
    def _eval(self, expr, n, extra_constraints=(), result=None, solver=None):
        global solve_count, cache_count

        #if n == 1 and model is None:
        #      import ipdb; ipdb.set_trace()

        results = [ ]
        model = result.backend_model if result else None
        if len(extra_constraints) > 0 or n != 1:
            solver.push()
        if len(extra_constraints) > 0:
            solver.add(*extra_constraints)
            model = None
            l.debug("Disregarding cache")

        for i in range(n):
            if model is None:
                solve_count += 1
                l.debug("Doing a check!")
                if solver.check() == z3.sat:
                    model = solver.model()
            else:
                cache_count += 1

            if model is None:
                break

            if not type(expr) in { int, float, str, bool, long }:
                v = model.eval(expr, model_completion=True)
            else:
                v = expr

            results.append(self._abstract(v))
            if i + 1 != n:
                solver.add(expr != v)
                model = None

        if len(extra_constraints) > 0 or n != 1:
            solver.pop()

        if len(results) == 0:
            raise UnsatError("constraints are unsat")

        return results

    @condom
    def _min(self, expr, extra_constraints=(), result=None, solver=None):
        global solve_count

        lo = 0
        hi = 2**expr.size()-1

        numpop = 0
        if len(extra_constraints) > 0:
            solver.push()
            numpop += 1
            solver.add(*[self.convert(e) for e in extra_constraints])

        while hi-lo > 1:
            middle = (lo + hi)/2
            #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)

            solver.push()
            solver.add(z3.UGE(expr, lo), z3.ULE(expr, middle))
            numpop += 1

            solve_count += 1
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                l.debug("... still sat")
                hi = middle
            else:
                l.debug("... now unsat")
                lo = middle
                solver.pop()
                numpop -= 1

        for _ in range(numpop):
            solver.pop()

        #l.debug("final hi/lo: %d, %d", hi, lo)

        if hi == lo: return NativeBVV(lo, expr.size())
        else:
            solver.push()
            solver.add(expr == lo)
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                solver.pop()
                return NativeBVV(lo, expr.size())
            else:
                solver.pop()
        return NativeBVV(hi, expr.size())

    @condom
    def _max(self, expr, extra_constraints=(), result=None, solver=None):
        global solve_count

        lo = 0
        hi = 2**expr.size()-1

        numpop = 0
        if len(extra_constraints) > 0:
            solver.push()
            numpop += 1
            solver.add(*[self.convert(e) for e in extra_constraints])

        while hi-lo > 1:
            middle = (lo + hi)/2
            #l.debug("h/m/l/d: %d %d %d %d", hi, middle, lo, hi-lo)

            solver.push()
            solver.add(z3.UGT(expr, middle), z3.ULE(expr, hi))
            numpop += 1

            solve_count += 1
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                l.debug("... still sat")
                lo = middle
            else:
                l.debug("... now unsat")
                hi = middle
                solver.pop()
                numpop -= 1
            #l.debug("        now: %d %d %d %d", hi, middle, lo, hi-lo)

        for _ in range(numpop):
            solver.pop()

        if hi == lo: return NativeBVV(lo, expr.size())
        else:
            solver.push()
            solver.add(expr == hi)
            l.debug("Doing a check!")
            if solver.check() == z3.sat:
                solver.pop()
                return NativeBVV(hi, expr.size())
            else:
                solver.pop()
        return NativeBVV(lo, expr.size())

    def _simplify(self, expr): #pylint:disable=W0613,R0201
        raise Exception("This shouldn't be called. Bug Yan.")

    @condom
    def simplify(self, expr):
        if expr._simplified:
            return expr

        if self._enable_simplification_cache:
            try:
                k = self._simplification_cache_key[expr._cache_key]
                #print "HIT WEAK KEY CACHE"
                return k
            except KeyError:
                pass
            try:
                k = self._simplification_cache_val[expr._cache_key]
                #print "HIT WEAK VALUE CACHE"
                return k
            except KeyError:
                pass

            #print "MISS CACHE"

        l.debug("SIMPLIFYING EXPRESSION")

        #print "SIMPLIFYING"

        expr_raw = self.convert(expr)

        #l.debug("... before: %s (%s)", expr_raw, expr_raw.__class__.__name__)

        if isinstance(expr_raw, z3.BoolRef):
            tactics = z3.Then(z3.Tactic("simplify"), z3.Tactic("propagate-ineqs"), z3.Tactic("propagate-values"), z3.Tactic("unit-subsume-simplify"))
            s = tactics(expr_raw).as_expr()
            n = s.decl().name()

            if n == 'true':
                s = True
            elif n == 'false':
                s = False
        elif isinstance(expr_raw, z3.BitVecRef):
            s = z3.simplify(expr_raw)
        else:
            s = expr_raw

        for b in _eager_backends:
            try:
                o = self.wrap(b.convert(s))
                break
            except BackendError:
                continue
        else:
            o = self._abstract(s)

        #print "SIMPLIFIED"
        #l.debug("... after: %s (%s)", s, s.__class__.__name__)

        o._simplified = Base.FULL_SIMPLIFY

        if self._enable_simplification_cache:
            self._simplification_cache_val[expr._cache_key] = o
            self._simplification_cache_key[expr._cache_key] = o
        return o

    @staticmethod
    def wrap(e):
        if isinstance(e, (z3.BitVecRef, NativeBVV)):
            return BVI(e, length=e.size())
        elif isinstance(e, (z3.BoolRef, bool)):
            return BoolI(e)
        else:
            raise Exception("whoops")

    def _is_false(self, e):
        return z3.simplify(e).eq(z3.BoolVal(False))

    def _is_true(self, e):
        return z3.simplify(e).eq(z3.BoolVal(True))

    def _solution(self, expr, v, result=None, extra_constraints=(), solver=None):
        return self._check(solver, extra_constraints=(expr == v,) + tuple(extra_constraints))

#
# this is for the actual->abstract conversion above
#

# these are Z3 operation names for abstraction
z3_op_names = [ _ for _ in dir(z3) if _.startswith('Z3_OP') ]
z3_op_nums = { getattr(z3, _): _ for _ in z3_op_names }

#pylint:disable=bad-continuation
op_map = {
    # Boolean
    'Z3_OP_TRUE': 'True',
    'Z3_OP_FALSE': 'False',
    'Z3_OP_EQ': '__eq__',
    'Z3_OP_DISTINCT': '__ne__',
    'Z3_OP_ITE': 'If',
    'Z3_OP_AND': 'And',
    'Z3_OP_OR': 'Or',
    'Z3_OP_IFF': '__eq__',
    'Z3_OP_XOR': 'Xor',
    'Z3_OP_NOT': 'Not',
    'Z3_OP_IMPLIES': 'Implies',
    #'Z3_OP_OEQ': None,

    # Arithmetic
    #'Z3_OP_ANUM': None,
    #'Z3_OP_AGNUM': None,
    'Z3_OP_LE': 'SLE',
    'Z3_OP_GE': 'SGE',
    'Z3_OP_LT': 'SLT',
    'Z3_OP_GT': 'SGT',
    'Z3_OP_ADD': '__add__',
    'Z3_OP_SUB': '__sub__',
    'Z3_OP_UMINUS': '__neg__',
    'Z3_OP_MUL': '__mul__',
    'Z3_OP_DIV': '__div__',
    'Z3_OP_IDIV': '__div__',
    'Z3_OP_REM': '__mod__', # TODO: is this correct?
    'Z3_OP_MOD': '__mod__',
    #'Z3_OP_TO_REAL': None,
    #'Z3_OP_TO_INT': None,
    #'Z3_OP_IS_INT': None,
    'Z3_OP_POWER': '__pow__',

    # Arrays & Sets
    #'Z3_OP_STORE': None,
    #'Z3_OP_SELECT': None,
    #'Z3_OP_CONST_ARRAY': None,
    #'Z3_OP_ARRAY_MAP': None,
    #'Z3_OP_ARRAY_DEFAULT': None,
    #'Z3_OP_SET_UNION': None,
    #'Z3_OP_SET_INTERSECT': None,
    #'Z3_OP_SET_DIFFERENCE': None,
    #'Z3_OP_SET_COMPLEMENT': None,
    #'Z3_OP_SET_SUBSET': None,
    #'Z3_OP_AS_ARRAY': None,

    # Bit-vectors
    'Z3_OP_BNUM': 'BitVecVal',
    #'Z3_OP_BIT1': None, # MAYBE TODO
    #'Z3_OP_BIT0': None, # MAYBE TODO
    'Z3_OP_BNEG': '__neg__',
    'Z3_OP_BADD': '__add__',
    'Z3_OP_BSUB': '__sub__',
    'Z3_OP_BMUL': '__mul__',

    'Z3_OP_BSDIV': '__div__',
    'Z3_OP_BUDIV': '__div__', # TODO: is this correct?
    'Z3_OP_BSREM': '__mod__', # TODO: is this correct?
    'Z3_OP_BUREM': '__mod__', # TODO: is this correct?
    'Z3_OP_BSMOD': '__mod__', # TODO: is this correct?

    # special functions to record the division by 0 cases
    # these are internal functions
    #'Z3_OP_BSDIV0': None,
    #'Z3_OP_BUDIV0': None,
    #'Z3_OP_BSREM0': None,
    #'Z3_OP_BUREM0': None,
    #'Z3_OP_BSMOD0': None,

    'Z3_OP_ULEQ': 'ULE',
    'Z3_OP_SLEQ': 'SLE',
    'Z3_OP_UGEQ': 'UGE',
    'Z3_OP_SGEQ': 'SGE',
    'Z3_OP_ULT': 'ULT',
    'Z3_OP_SLT': 'SLT',
    'Z3_OP_UGT': 'UGT',
    'Z3_OP_SGT': 'SGT',

    'Z3_OP_BAND': '__and__',
    'Z3_OP_BOR': '__or__',
    'Z3_OP_BNOT': '__invert__',
    'Z3_OP_BXOR': '__xor__',
    #'Z3_OP_BNAND': None,
    #'Z3_OP_BNOR': None,
    #'Z3_OP_BXNOR': None,

    'Z3_OP_CONCAT': 'Concat',
    'Z3_OP_SIGN_EXT': 'SignExt',
    'Z3_OP_ZERO_EXT': 'ZeroExt',
    'Z3_OP_EXTRACT': 'Extract',
    'Z3_OP_REPEAT': 'RepeatBitVec',

    #'Z3_OP_BREDOR': None,
    #'Z3_OP_BREDAND': None,
    #'Z3_OP_BCOMP': None,

    'Z3_OP_BSHL': '__lshift__',
    'Z3_OP_BLSHR': 'LShR',
    'Z3_OP_BASHR': '__rshift__',
    #'Z3_OP_ROTATE_LEFT': None,
    #'Z3_OP_ROTATE_RIGHT': None,
    'Z3_OP_EXT_ROTATE_LEFT': 'RotateLeft',
    'Z3_OP_EXT_ROTATE_RIGHT': 'RotateRight',

    'Z3_OP_FPA_TO_SBV': 'fpToSBV',
    'Z3_OP_FPA_TO_UBV': 'fpToUBV',
    'Z3_OP_FPA_TO_IEEE_BV': 'fpToIEEEBV',
    'Z3_OP_FPA_TO_FP': 'fpToFP',
    'Z3_OP_FPA_NUM': 'FPVal',

    'Z3_OP_FPA_MINUS_ZERO': 'MinusZero',
    'Z3_OP_FPA_MINUS_INF': 'MinusInf',
    'Z3_OP_FPA_PLUS_ZERO': 'PlusZero',
    'Z3_OP_FPA_PLUS_INF': 'PlusInf',
    'Z3_OP_FPA_NAN': 'NaN',

    'Z3_OP_FPA_EQ': 'fpEQ',
    'Z3_OP_FPA_GT': 'fpGT',
    'Z3_OP_FPA_GE': 'fpGEQ',
    'Z3_OP_FPA_LT': 'fpLT',
    'Z3_OP_FPA_LE': 'fpLEQ',

    'Z3_OP_FPA_ABS': 'fpAbs',
    'Z3_OP_FPA_NEG': 'fpNeg',
    'Z3_OP_FPA_ADD': 'fpAdd',
    'Z3_OP_FPA_SUB': 'fpSub',
    'Z3_OP_FPA_MUL': 'fpMul',
    'Z3_OP_FPA_DIV': 'fpDiv',

    'Z3_OP_FPA_RM_NEAREST_TIES_TO_EVEN': 'RM_RNE',
    'Z3_OP_FPA_RM_NEAREST_TIES_TO_AWAY': 'RM_RNA',
    'Z3_OP_FPA_RM_TOWARD_ZERO': 'RM_RTZ',
    'Z3_OP_FPA_RM_TOWARD_POSITIVE': 'RM_RTP',
    'Z3_OP_FPA_RM_TOWARD_NEGATIVE': 'RM_RTN',

    'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
}

from ..ast.base import Base, model_object
from ..ast.bv import BV, BVI
from ..ast.bool import BoolI, Bool
from ..ast.fp import FP, FPI
from ..operations import backend_operations, backend_fp_operations, bin_ops
from ..result import Result
from ..bv import BVV as NativeBVV
from ..fp import FPV as NativeFPV, FSort, RM, RM_RNE, RM_RNA, RM_RTP, RM_RTN, RM_RTZ
from ..errors import ClaripyError, BackendError, UnsatError, ClaripyOperationError
from .. import _eager_backends, _all_operations

op_type_map = {
    # Boolean
    'Z3_OP_TRUE': Bool,
    'Z3_OP_FALSE': Bool,
    'Z3_OP_EQ': Bool,
    'Z3_OP_DISTINCT': Bool,
    'Z3_OP_ITE': Bool,
    'Z3_OP_AND': Bool,
    'Z3_OP_OR': Bool,
    'Z3_OP_IFF': Bool,
    'Z3_OP_XOR': Bool,
    'Z3_OP_NOT': Bool,
    'Z3_OP_IMPLIES': Bool,
    #'Z3_OP_OEQ': None,

    # Arithmetic
    #'Z3_OP_ANUM': None,
    #'Z3_OP_AGNUM': None,
    'Z3_OP_LE': None,
    'Z3_OP_GE': None,
    'Z3_OP_LT': None,
    'Z3_OP_GT': None,
    'Z3_OP_ADD': None,
    'Z3_OP_SUB': None,
    'Z3_OP_UMINUS': None,
    'Z3_OP_MUL': None,
    'Z3_OP_DIV': None,
    'Z3_OP_IDIV': None,
    'Z3_OP_REM': None, # TODO: is this correct?
    'Z3_OP_MOD': None,
    #'Z3_OP_TO_REAL': None,
    #'Z3_OP_TO_INT': None,
    #'Z3_OP_IS_INT': None,
    'Z3_OP_POWER': None,

    # Arrays & Sets
    #'Z3_OP_STORE': None,
    #'Z3_OP_SELECT': None,
    #'Z3_OP_CONST_ARRAY': None,
    #'Z3_OP_ARRAY_MAP': None,
    #'Z3_OP_ARRAY_DEFAULT': None,
    #'Z3_OP_SET_UNION': None,
    #'Z3_OP_SET_INTERSECT': None,
    #'Z3_OP_SET_DIFFERENCE': None,
    #'Z3_OP_SET_COMPLEMENT': None,
    #'Z3_OP_SET_SUBSET': None,
    #'Z3_OP_AS_ARRAY': None,

    # Bit-vectors
    'Z3_OP_BNUM': 'BitVecVal',
    #'Z3_OP_BIT1': None, # MAYBE TODO
    #'Z3_OP_BIT0': None, # MAYBE TODO
    'Z3_OP_BNEG': BV,
    'Z3_OP_BADD': BV,
    'Z3_OP_BSUB': BV,
    'Z3_OP_BMUL': BV,

    'Z3_OP_BSDIV': BV,
    'Z3_OP_BUDIV': BV,
    'Z3_OP_BSREM': BV,
    'Z3_OP_BUREM': BV,
    'Z3_OP_BSMOD': BV,

    # special functions to record the division by 0 cases
    # these are internal functions
    #'Z3_OP_BSDIV0': None,
    #'Z3_OP_BUDIV0': None,
    #'Z3_OP_BSREM0': None,
    #'Z3_OP_BUREM0': None,
    #'Z3_OP_BSMOD0': None,

    'Z3_OP_ULEQ': Bool,
    'Z3_OP_SLEQ': Bool,
    'Z3_OP_UGEQ': Bool,
    'Z3_OP_SGEQ': Bool,
    'Z3_OP_ULT': Bool,
    'Z3_OP_SLT': Bool,
    'Z3_OP_UGT': Bool,
    'Z3_OP_SGT': Bool,

    'Z3_OP_BAND': BV,
    'Z3_OP_BOR': BV,
    'Z3_OP_BNOT': BV,
    'Z3_OP_BXOR': BV,
    #'Z3_OP_BNAND': None,
    #'Z3_OP_BNOR': None,
    #'Z3_OP_BXNOR': None,

    'Z3_OP_CONCAT': BV,
    'Z3_OP_SIGN_EXT': BV,
    'Z3_OP_ZERO_EXT': BV,
    'Z3_OP_EXTRACT': BV,
    'Z3_OP_REPEAT': BV,

    #'Z3_OP_BREDOR': None,
    #'Z3_OP_BREDAND': None,
    #'Z3_OP_BCOMP': None,

    'Z3_OP_BSHL': BV,
    'Z3_OP_BLSHR': BV,
    'Z3_OP_BASHR': BV,
    #'Z3_OP_ROTATE_LEFT': None,
    #'Z3_OP_ROTATE_RIGHT': None,
    'Z3_OP_EXT_ROTATE_LEFT': BV,
    'Z3_OP_EXT_ROTATE_RIGHT': BV,

    'Z3_OP_FPA_TO_SBV': BV,
    'Z3_OP_FPA_TO_UBV': BV,
    'Z3_OP_FPA_TO_IEEE_BV': BV,
    'Z3_OP_FPA_TO_FP': FP,
    'Z3_OP_FPA_NUM': FP,

    'Z3_OP_FPA_MINUS_ZERO': FP,
    'Z3_OP_FPA_MINUS_INF': FP,
    'Z3_OP_FPA_PLUS_ZERO': FP,
    'Z3_OP_FPA_PLUS_INF': FP,
    'Z3_OP_FPA_NAN': FP,

    'Z3_OP_FPA_EQ': Bool,
    'Z3_OP_FPA_GT': Bool,
    'Z3_OP_FPA_GE': Bool,
    'Z3_OP_FPA_LT': Bool,
    'Z3_OP_FPA_LE': Bool,

    'Z3_OP_FPA_ABS': FP,
    'Z3_OP_FPA_NEG': FP,
    'Z3_OP_FPA_ADD': FP,
    'Z3_OP_FPA_SUB': FP,
    'Z3_OP_FPA_MUL': FP,
    'Z3_OP_FPA_DIV': FP,

    'Z3_OP_UNINTERPRETED': 'UNINTERPRETED'
}



import logging
l = logging.getLogger("claripy.backends.backend_concrete")

import z3

from ..backend import BackendError, Backend

class BackendConcrete(Backend):
    def __init__(self):
        Backend.__init__(self)
        self._make_raw_ops(set(backend_operations) - { 'BitVec', 'If' }, op_module=bv)
        self._make_raw_ops(backend_fp_operations, op_module=fp)
        self._op_raw_result['BitVec'] = self._BitVec
        self._op_raw_result['If'] = self._If
        self._z_true = z3.BoolVal(True)
        self._z_false = z3.BoolVal(False)

    def _BitVec(self, name, size, result=None): #pylint:disable=W0613,R0201
        if result is None:
            l.debug("BackendConcrete can only handle BitVec when we are given a model")
            raise BackendError("BackendConcrete can only handle BitVec when we are given a model")
        if name not in result.model:
            l.debug("BackendConcrete needs variable %s in the model", name)
            raise BackendError("BackendConcrete needs variable %s in the model" % name)
        else:
            return result.model[name]

    def _If(self, b, t, f, result=None): #pylint:disable=no-self-use,unused-argument
        if not isinstance(b, bool):
            raise BackendError("BackendConcrete can't handle non-bool condition in If.")
        else:
            return t if b else f

    def _size(self, e, result=None):
        if type(e) in { bool, long, int }:
            return None
        elif type(e) in { bv.BVV }:
            return e.size()
        elif isinstance(e, fp.FPV):
            return e.sort.length
        else:
            raise BackendError("can't get size of type %s" % type(e))

    def _name(self, e, result=None): #pylint:disable=unused-argument,no-self-use
        return None

    def _identical(self, a, b, result=None):
        if type(a) is bv.BVV and type(b) is bv.BVV and a.size() != b.size():
            return False
        else:
            return a == b

    def _convert(self, a, result=None):
        if type(a) in { int, long, float, bool, str, bv.BVV, fp.FPV, fp.RM, fp.FSort }:
            return a

        if not hasattr(a, '__module__') or a.__module__ != 'z3':
            raise BackendError("BackendConcrete got an unsupported type %s" % a.__class__)

        #if _backend_z3 is None:
        #   raise BackendError("can't convert z3 expressions when z3 is not in use")

        try:
            #if hasattr(_backend_z3, '_lock'):
            #   _backend_z3._lock.acquire() #pylint:disable=no-member

            if hasattr(a, 'as_long'): return bv.BVV(a.as_long(), a.size())
            elif isinstance(a, z3.FPNumRef):
                # TODO: don't replicate this code in backend_z3.py
                # this is really imprecise
                fp_mantissa = float(a.significand())
                fp_exp = float(a.exponent())
                value = fp_mantissa * (2 ** fp_exp)

                ebits = a.ebits()
                sbits = a.sbits()
                sort = fp.FSort.from_params(ebits, sbits)

                return fp.FPV(value, sort)
            elif isinstance(a, z3.BoolRef) and a.eq(self._z_true): return True
            elif isinstance(a, z3.BoolRef) and a.eq(self._z_false): return False
            elif result is not None and a.num_args() == 0:
                name = a.decl().name()
                if name in result.model:
                    return result.model[name]
                else:
                    l.debug("returning 0 for %s (not in model)", name)
                    return bv.BVV(0, a.size())
            else:
                #import ipdb; ipdb.set_trace()
                #l.warning("TODO: support more complex non-symbolic expressions, maybe?")
                raise BackendError("TODO: support more complex non-symbolic expressions, maybe?")
        finally:
            pass
            #if hasattr(_backend_z3, '_lock'):
            #   _backend_z3._lock.release() #pylint:disable=no-member

    def _simplify(self, e):
        return e

    def abstract(self, e):
        if isinstance(e, bv.BVV):
            return BVI(e, length=e.size())
        elif isinstance(e, bool):
            return BoolI(e)
        elif isinstance(e, fp.FPV):
            return FPI(e)
        else:
            raise BackendError("Couldn't abstract object of type {}".format(type(e)))

    #
    # Evaluation functions
    #

    def _eval(self, expr, n, result=None, solver=None, extra_constraints=()):
        return (expr,)
    def _max(self, expr, result=None, solver=None, extra_constraints=()):
        return expr
    def _min(self, expr, result=None, solver=None, extra_constraints=()):
        return expr
    def _solution(self, expr, v, result=None, solver=None, extra_constraints=()):
        return self.convert(expr, result=result) == v

    def _is_true(self, e):
        return e == True
    def _is_false(self, e):
        return e == False
    def _has_true(self, e):
        return e == True
    def _has_false(self, e):
        return e == False

from ..operations import backend_operations, backend_fp_operations
from .. import bv, fp
#from .. import _backend_z3
from ..ast.bv import BVI
from ..ast.fp import FPI
from ..ast.bool import BoolI



import os
import sys
import struct
import weakref
import hashlib
import itertools
import cPickle as pickle

import logging
l = logging.getLogger("claripy.ast")

import ana

if os.environ.get('WORKER', False):
    WORKER = True
else:
    WORKER = False

md5_unpacker = struct.Struct('2Q')

#pylint:enable=unused-argument
#pylint:disable=unidiomatic-typecheck

def _is_eager(a):
    if isinstance(a, (int, long, bool, RM, FSort)):
        return True
    elif isinstance(a, Base):
        return a.eager
    else:
        return False

def _inner_repr(a, **kwargs):
    if isinstance(a, Base):
        return a.__repr__(inner=True, **kwargs)
    else:
        return repr(a)

class ASTCacheKey(object): pass

#
# AST variable naming
#

var_counter = itertools.count()
_unique_names = True

def _make_name(name, size, explicit_name=False, prefix=""):
    if _unique_names and not explicit_name:
        return "%s%s_%d_%d" % (prefix, name, var_counter.next(), size)
    else:
        return name


class Base(ana.Storable):
    '''
    An AST tracks a tree of operations on arguments. It has the following methods:

        op: the operation that is being done on the arguments
        args: the arguments that are being used
        length: the length (in bits)

    AST objects have *hash identity*. This means that an AST that has the same hash as
    another AST will be the *same* object. For example, the following is true:

        a, b = two different ASTs
        c = b + a
        d = b + a
        assert c is d

    This is done to better support serialization and better manage memory.
    '''

    __slots__ = [ 'op', 'args', 'variables', 'symbolic', '_objects', '_collapsible', '_hash', '_simplified', '_cache_key', '_errored', 'eager', 'length' ]
    _hash_cache = weakref.WeakValueDictionary()

    FULL_SIMPLIFY=1
    LITE_SIMPLIFY=2
    UNSIMPLIFIED=0

    def __new__(cls, op, args, **kwargs):
        '''
        This is called when you create a new Base object, whether directly or through an operation.
        It finalizes the arguments (see the _finalize function, above) and then computes
        a hash. If an AST of this hash already exists, it returns that AST. Otherwise,
        it creates, initializes, and returns the AST.

        @param op: the AST operation ('__add__', 'Or', etc)
        @param args: the arguments to the AST operation (i.e., the objects to add)
        @param variables: the symbolic variables present in the AST (default: empty set)
        @param symbolic: a flag saying whether or not the AST is symbolic (default: False)
        @param length: an integer specifying the length of this AST (default: None)
        @param collapsible: a flag of whether or not Claripy can feel free to collapse this AST.
                            This is mostly used to keep Claripy from collapsing Reverse operations,
                            so that they can be undone with another Reverse.
        @param simplified: a measure of how simplified this AST is. 0 means unsimplified, 1 means
                           fast-simplified (basically, just undoing the Reverse op), and 2 means
                           simplified through z3.
        @param errored: a set of backends that are known to be unable to handle this AST.
        @param eager: whether or not to evaluate future parent ASTs eagerly.
        '''
        if any(not isinstance(a, (str, int, long, bool, Base, BackendObject, FSort)) for a in args):
            #import ipdb; ipdb.set_trace()
            raise ClaripyTypeError("arguments %s contain an unknown type to claripy.Base" % (args,))

        # fix up args and kwargs
        a_args = tuple((a.to_claripy() if isinstance(a, BackendObject) else a) for a in args)
        if 'symbolic' not in kwargs:
            kwargs['symbolic'] = any(a.symbolic for a in a_args if isinstance(a, Base))
        if 'variables' not in kwargs:
            kwargs['variables'] = frozenset.union(
                frozenset(), *(a.variables for a in a_args if isinstance(a, Base))
            )
        elif type(kwargs['variables']) is not frozenset: #pylint:disable=unidiomatic-typecheck
            kwargs['variables'] = frozenset(kwargs['variables'])
        if 'errored' not in kwargs:
            kwargs['errored'] = set.union(set(), *(a._errored for a in a_args if isinstance(a, Base)))

        if 'add_variables' in kwargs:
            kwargs['variables'] = kwargs['variables'] | kwargs['add_variables']

        h = Base._calc_hash(op, a_args, kwargs)

        self = cls._hash_cache.get(h, None)
        if self is None:
            self = super(Base, cls).__new__(cls, op, a_args, **kwargs)
            self.__a_init__(op, a_args, **kwargs)
            self._hash = h
            cls._hash_cache[h] = self
        # else:
        #    if self.args != f_args or self.op != f_op or self.variables != f_kwargs['variables']:
        #        raise Exception("CRAP -- hash collision")

        return self

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def _calc_hash(op, args, k):
        '''
        Calculates the hash of an AST, given the operation, args, and kwargs.

        @param op: the operation
        @param args: the arguments to the operation
        @param kwargs: a dict including the 'symbolic', 'variables', and 'length' items

        @returns a hash
        '''
        to_hash = (op, tuple(str(a) if type(a) in (int, long) else hash(a) for a in args), k['symbolic'], hash(k['variables']), str(k.get('length', None)))
        # Why do we use md5 when it's broken? Because speed is more important
        # than cryptographic integrity here. Then again, look at all those
        # allocations we're doing here... fast python is painful.
        hd = hashlib.md5(pickle.dumps(to_hash, -1)).digest()
        return md5_unpacker.unpack(hd)[0] # 64 bits

    #pylint:disable=attribute-defined-outside-init
    def __a_init__(self, op, args, variables=None, symbolic=None, length=None, collapsible=None, simplified=0, errored=None, eager=False, add_variables=None): #pylint:disable=unused-argument
        '''
        Initializes an AST. Takes the same arguments as Base.__new__()
        '''
        self.op = op
        self.args = args
        self.length = length
        self.variables = frozenset(variables)
        self.symbolic = symbolic
        self.eager = eager

        self._collapsible = True if collapsible is None else collapsible
        self._errored = errored if errored is not None else set()

        self._simplified = simplified
        self._cache_key = ASTCacheKey()

        if self.op != 'I' and all(_is_eager(a) for a in self.args):
            model = self.model
            if model is not self:
                self.op = 'I'
                self.args = (model,)

                # Usually `eagerness` should be passed on. However, type of the model might be different than its
                # arguments. In VSA, for instance, a union of two BVVs can lead to a StridedInterval instance, where
                # BVVs should be evaluated eagerly while StridedIntervals should not be, Hence we are rechecking if the
                # model itself should be eagerly evaluated and property set the eager property here.
                self.eager = _is_eager(model)

        if len(args) == 0:
            raise ClaripyOperationError("AST with no arguments!")

        #if self.op != 'I':
        #    for a in args:
        #        if not isinstance(a, Base) and type(a) not in (int, long, bool, str, unicode):
        #            import ipdb; ipdb.set_trace()
        #            l.warning(ClaripyOperationError("Un-wrapped native object of type %s!" % type(a)))
    #pylint:enable=attribute-defined-outside-init

    def make_uuid(self, uuid=None):
        '''
        This overrides the default ANA uuid with the hash of the AST. UUID is slow,
        and we'll soon replace it from ANA itself, and this will go away.

        @returns a string representation of the AST hash.
        '''
        u = getattr(self, '_ana_uuid', None)
        if u is None:
            u = str(self._hash) if uuid is None else uuid
            ana.get_dl().uuid_cache[u] = self
            setattr(self, '_ana_uuid', u)
        return u

    @property
    def uuid(self):
        return self.ana_uuid

    def __hash__(self):
        return self._hash

    #
    # Serialization support
    #

    def _ana_getstate(self):
        '''
        Support for ANA serialization.
        '''
        return self.op, self.args, self.length, self.variables, self.symbolic, self._hash
    def _ana_setstate(self, state):
        '''
        Support for ANA deserialization.
        '''
        op, args, length, variables, symbolic, h = state
        Base.__a_init__(self, op, args, length=length, variables=variables, symbolic=symbolic)
        self._hash = h
        Base._hash_cache[h] = self

    #
    # Collapsing and simplification
    #

    #def _models_for(self, backend):
    #    for a in self.args:
    #        backend.convert_expr(a)
    #        else:
    #            yield backend.convert(a)

    def make_like(self, *args, **kwargs):
        return type(self)(*args, **kwargs)

    def _should_collapse(self):
        '''
        This is a helper function that checks if the AST is "collapsible". It returns
        False if there is some reason to keep it from being collapsed. For example,
        an AST that contains only one StridedInterval in its immediate arguments should
        not be collapsed, because the VSA does tricky things to avoid losing precision.
        '''
        raw_args = self.arg_models()
        types = [ type(a) for a in raw_args ]

        #l.debug("In should_collapse()")

        if types.count(Base) != 0 and not all((a._collapsible for a in raw_args if isinstance(a, Base))):
                #l.debug("... not collapsing for op %s because ASTs are present.", self.op)
                return False

        if self.op in operations.not_invertible:
            #l.debug("... collapsing the AST for operation %s because it's not invertible", self.op)
            return True

        constants = sum((types.count(t) for t in (int, long, bool, str, bv.BVV)))
        if constants == len(raw_args):
            #l.debug("... collapsing the AST for operation %s because it's full of constants", self.op)
            return True

        if len([ a for a in raw_args if isinstance(a, StridedInterval) and a.is_integer]) > 1:
            #l.debug("... collapsing the AST for operation %s because there are more than two SIs", self.op)
            return True

        #
        # More complex checks probably go here.
        #

        # Reversible; don't collapse!
        #l.debug("not collapsing the AST for operation %s!", self.op)
        return False

    @property
    def collapsed(self):
        '''
        A collapsed version of the AST, if the AST *can* be collapsed.
        '''
        if self._should_collapse() and self._collapsible:
            #l.debug("Collapsing!")
            r = self.model
            if not isinstance(r, Base):
                if isinstance(self, Bits):
                    return self.__class__('I', args=(r,), length=len(self), variables=self.variables, symbolic=self.symbolic)
                else:
                    return self.__class__('I', args=(r,), variables=self.variables, symbolic=self.symbolic)
            else:
                return r
        else:
            return self

    @property
    def simplified(self):
        '''
        A lightly simplified version of the AST (simplify level 1). This basically
        just cancels out two Reverse operations, if they are present. Later on, it will hopefully
        do more.
        '''

        if self._simplified:
            return self

        if self.op in operations.reverse_distributable and all((isinstance(a, Base) for a in self.args)) and set((a.op for a in self.args)) == { 'Reverse' }:
            inner_a = self.make_like(self.op, tuple(a.args[0] for a in self.args)).simplified
            o = self.make_like('Reverse', (inner_a,), collapsible=True).simplified
            o._simplified = Base.LITE_SIMPLIFY
            return o

        # self = self.make_like(self._claripy, self.op, tuple(a.reduced if isinstance(a, Base) else a for a in self.args))

        return self

    @property
    def reduced(self):
        '''
        A simplified, collapsed version of the AST.
        '''
        a = self.simplified
        if isinstance(a, Base):
            return a.collapsed
        else:
            return a

    # No more size in Base

    @property
    def cardinality(self):
        t = type(self.model)

        if t in (int, long, bool, str, bv.BVV):
            return 1

        elif t in (vsa.IfProxy, vsa.StridedInterval, vsa.ValueSet, vsa.DiscreteStridedIntervalSet):
            return self.model.cardinality

        else:
            raise NotImplementedError("'cardinality' is not supported in modes other than static mode")

    #
    # Functionality for resolving to model objects
    #

    def arg_models(self):
        '''
        Helper function to return the model (i.e., non-AST) objects of the arguments.
        '''
        return [ (a.model if isinstance(a, Base) else a) for a in self.args ]

    def resolved(self, result=None):
        '''
        Returns a model object (i.e., an object that is the result of all the operations
        in the AST), if there is a backend that can handle this AST. Otherwise, return
        itself.

        @arg result: a Result object, for resolving symbolic variables using the
                     concrete backend.
        '''
        for b in _model_backends:
            try: return self.resolved_with(b, result=result)
            except BackendError: self._errored.add(b)
        l.debug("all model backends failed for op %s", self.op)
        return self

    @property
    def model(self):
        '''
        The model object (the result of the operation represented by this AST).
        '''
        r = self.resolved()
        return r

    def resolved_with(self, b, result=None):
        '''
        Returns the result of carrying out the operation of this AST with the
        specified backend.

        @arg b: the backend to resolve with
        @arg result: a Result object, for resolving symbolic variables using the
                     concrete backend
        '''
        if b in self._errored and result is None:
            raise BackendError("%s already failed" % b)

        #l.debug("trying evaluation with %s", b)
        return b.call(self, result=result)

    #
    # Viewing and debugging
    #

    def dbg_repr(self, prefix=None):
        try:
            if prefix is not None:
                new_prefix = prefix + "    "
                s = prefix + "<%s %s (\n" % (type(self).__name__, self.op)
                for a in self.args:
                    s += "%s,\n" % (a.dbg_repr(prefix=new_prefix) if hasattr(a, 'dbg_repr') else (new_prefix + repr(a)))
                s = s[:-2] + '\n'
                s += prefix + ")>"

                return s
            else:
                return "<%s %s (%s)>" % (type(self).__name__, self.op, ', '.join(a.dbg_repr() if hasattr(a, 'dbg_repr') else repr(a) for a in self.args))
        except RuntimeError:
            e_type, value, traceback = sys.exc_info()
            raise ClaripyRecursionError, ("Recursion limit reached during display. I sorry.", e_type, value), traceback

    def _type_name(self):
        return self.__class__.__name__

    def __repr__(self, inner=False, explicit_length=False):
        if WORKER:
            return '<AST something>'

        self = self.reduced

        if not isinstance(self.model, Base):
            if inner:
                if isinstance(self.model, bv.BVV):
                    if self.model.value < 10:
                        val = format(self.model.value, '')
                    else:
                        val = format(self.model.value, '#x')
                    return val + ('#' + str(self.model.bits) if explicit_length else '')
                else:
                    return repr(self.model)
            else:
                return '<{} {}>'.format(self._type_name(), self.model)
        else:
            try:
                if self.op in operations.reversed_ops:
                    op = operations.reversed_ops[self.op]
                    args = self.args[::-1]
                else:
                    op = self.op
                    args = self.args

                if op == 'BitVec' and inner:
                    value = args[0]
                elif op == 'If':
                    value = 'if {} then {} else {}'.format(_inner_repr(args[0]),
                                                           _inner_repr(args[1]),
                                                           _inner_repr(args[2]))
                    if inner:
                        value = '({})'.format(value)
                elif op == 'Not':
                    value = '!{}'.format(_inner_repr(args[0]))
                elif op == 'Extract':
                    value = '{}[{}:{}]'.format(_inner_repr(args[2]), args[0], args[1])
                elif op == 'ZeroExt':
                    value = '0#{} .. {}'.format(args[0], _inner_repr(args[1]))
                    if inner:
                        value = '({})'.format(value)
                elif op == 'Concat':
                    value = ' .. '.join(_inner_repr(a, explicit_length=True) for a in self.args)
                elif len(args) == 2 and op in operations.infix:
                    value = '{} {} {}'.format(_inner_repr(args[0]),
                                              operations.infix[op],
                                              _inner_repr(args[1]))
                    if inner:
                        value = '({})'.format(value)
                else:
                    value = "{}({})".format(op,
                                            ', '.join(_inner_repr(a) for a in args))

                if not inner:
                    value = '<{} {}>'.format(self._type_name(), value)

                return value
            except RuntimeError:
                e_type, value, traceback = sys.exc_info()
                raise ClaripyRecursionError, ("Recursion limit reached during display. I sorry.", e_type, value), traceback

    @property
    def depth(self):
        '''
        The depth of this AST. For example, an AST representing (a+(b+c)) would have
        a depth of 2.
        '''
        if self.op == 'BitVec':
            return 0
        ast_args = [ a for a in self.args if isinstance(a, Base) ]
        return 1 + (max(a.depth for a in ast_args) if len(ast_args) > 0 else 1)

    @property
    def recursive_children_asts(self):
        for a in self.args:
            if isinstance(a, Base):
                l.debug("Yielding AST %s with hash %s with %d children", a, hash(a), len(a.args))
                yield a
                for b in a.recursive_children_asts:
                    yield b

    def dbg_is_looped(self, seen=None, checked=None):
        seen = set() if seen is None else seen
        checked = set() if checked is None else checked

        seen.add(hash(self))
        for a in self.recursive_children_asts:
            l.debug("Checking AST with hash %s for looping", hash(a))
            if hash(a) in seen:
                return a
            elif hash(a) in checked:
                return False
            else:
                for a in self.args:
                    if not isinstance(a, Base):
                        continue

                    r = a.dbg_is_looped(seen=seen, checked=checked)
                    if r is not False: return r
                    else: checked.add(hash(a))

        seen.discard(hash(self))
        return False

    #
    # Various AST modifications (replacements)
    #

    def _replace(self, old, new, replacements=None):
        '''
        A helper for replace().
        '''
        if replacements is None:
            replacements = { hash(old): new }

        hash_key = hash(self)

        if hash_key in replacements:
            r = replacements[hash_key]
        elif not self.variables.issuperset(old.variables):
            r = self
        else:
            new_args = [ ]
            replaced = False

            for a in self.args:
                if isinstance(a, Base):
                    new_a = a._replace(old, new, replacements=replacements)
                    replaced |= hash(new_a) != hash(a)
                else:
                    new_a = a

                new_args.append(new_a)

            if replaced:
                r = self.make_like(self.op, tuple(new_args)).reduced
            else:
                r = self

        replacements[hash_key] = r
        return r


    #
    # Other helper functions
    #

    def split(self, split_on):
        '''
        Splits the AST if its operation is split_on (i.e., return all the arguments).
        Otherwise, return a list with just the AST.
        '''
        if self.op in split_on: return list(self.args)
        else: return [ self ]

    # we don't support iterating over Base objects
    def __iter__(self):
        '''
        This prevents people from iterating over ASTs.
        '''
        raise ClaripyOperationError("Please don't iterate over, or split, AST nodes!")

    def __nonzero__(self):
        '''
        This prevents people from accidentally using an AST as a condition. For
        example, the following was previously common:

            a,b = two ASTs
            if a == b:
                do something

        The problem is that `a == b` would return an AST, because an AST can be symbolic
        and there could be no way to actually know the value of that without a
        constraint solve. This caused tons of issues.
        '''
        raise ClaripyOperationError('testing Expressions for truthiness does not do what you want, as these expressions can be symbolic')

    def identical(self, o):
        '''
        Returns True if 'o' can be easily determined to be identical to this AST.
        Otherwise, return False. Note that the AST *might* still be identical (i.e.,
        if it were simplified via Z3), but it's hard to quickly tell that.
        '''
        return is_identical(self, o)

    def structurally_match(self, o):
        """
        Structurally compares two A objects, and check if their corresponding leaves are definitely the same A object
        (name-wise or hash-identity wise).

        :param o: the other claripy A object
        :return: True/False
        """

        # TODO: Convert a and b into canonical forms

        if self.op != o.op:
            return False

        if len(self.args) != len(o.args):
            return False

        for arg_a, arg_b in zip(self.args, o.args):
            if not isinstance(arg_a, Base):
                if type(arg_a) != type(arg_b):
                    return False
                # They are not ASTs
                if arg_a != arg_b:
                    return False
                else:
                    continue

            if arg_a.op in ('I', 'BitVec', 'FP'):
                # This is a leaf node in AST tree
                if arg_a is not arg_b:
                    return False

            else:
                if not arg_a.structurally_match(arg_b):
                    return False

        return True

    def replace(self, old, new):
        '''
        Returns an AST with all instances of the AST 'old' replaced with AST 'new'
        '''
        if not isinstance(old, Base) or not isinstance(new, Base):
            raise ClaripyOperationError('replacements must be AST nodes')
        if not isinstance(new, Bool):
            if old.size() != new.size():
                raise ClaripyOperationError('replacements must have matching sizes')
        return self._replace(old, new)

    def _identify_vars(self, all_vars, counter):
        if self.op == 'BitVec':
            if self.args not in all_vars:
                all_vars[self.args] = BV('var_' + str(next(counter)),
                                         self.args[1],
                                         explicit_name=True)
        else:
            for arg in self.args:
                if isinstance(arg, Base):
                    arg._identify_vars(all_vars, counter)

    def canonicalized(self, existing_vars=None, counter=None):
        all_vars = {} if existing_vars is None else existing_vars
        counter = itertools.count() if counter is None else counter
        self._identify_vars(all_vars, counter)

        expr = self
        for old_var, new_var in all_vars.items():
            expr = expr.replace(BV(*old_var, explicit_name=True), new_var)

        return all_vars, expr

#
# Unbound methods
#

def is_identical(*args):
    '''
    Attempts to check if the underlying models of the expression are identical,
    even if the hashes match.

    This process is somewhat conservative: False does not necessarily mean that
    it's not identical; just that it can't (easily) be determined to be identical.
    '''
    if not all([isinstance(a, Base) for a in args]):
        return False

    if len(set(hash(a) for a in args)) == 1:
        return True

    first = args[0]
    identical = None
    for o in args:
        for b in _all_backends:
            try:
                i = b.identical(first, o)
                if identical is None:
                    identical = True
                identical &= i is True
            except BackendError:
                pass

        if not identical:
            return False

    return identical is True

def model_object(e, result=None):
    for b in _all_backends:
        try: return b.convert(e, result=result)
        except BackendError: pass
    raise ClaripyTypeError('no model backend can convert expression')

def simplify(e):
    if isinstance(e, Base) and e.op == 'I':
        return e

    for b in _all_backends:
        try:
            s = b.simplify(e)
            s._simplified = Base.FULL_SIMPLIFY
            return s
        except BackendError:
            pass

    l.debug("Unable to simplify expression")
    return e

from ..errors import BackendError, ClaripyOperationError, ClaripyRecursionError, ClaripyTypeError
from .. import operations
from .. import bv, vsa
from ..fp import RM, FSort
from ..vsa import StridedInterval
from ..backend_object import BackendObject
from .. import _all_backends, _model_backends
from ..ast.bits import Bits
from ..ast.bool import Bool
from ..ast.bv import BV



import copy
import collections
import weakref

import threading

class Result(object):
    def __init__(self, satness, model=None, approximation=False, backend_model=None):
        self.sat = satness
        self.model = model if model is not None else { }
        self._tls = threading.local()
        self._tls.backend_model = backend_model
        self.approximation = approximation
        self.eval_cache = { }
        self.min_cache = { }
        self.max_cache = { }
        self.resolve_cache = collections.defaultdict(weakref.WeakKeyDictionary)

    @property
    def backend_model(self):
        try:
            return self._tls.backend_model
        except AttributeError:
            return None

    def branch(self):
        r = Result(self.sat, copy.copy(self.model), backend_model=self._tls.backend_model)
        r.eval_cache = dict(self.eval_cache)
        r.min_cache = dict(self.min_cache)
        r.max_cache = dict(self.max_cache)
        return r

    def __getstate__(self):
        return ( self.sat, self.model, self.eval_cache, self.min_cache, self.max_cache )

    def __setstate__(self, state):
        ( self.sat, self.model, self.eval_cache, self.min_cache, self.max_cache ) = state
        self.resolve_cache = collections.defaultdict(weakref.WeakKeyDictionary)
        self._tls = threading.local()
        self._tls.backend_model = None

    def downsize(self):
        self._tls.backend_model = None

def UnsatResult(**kwargs):
    return Result(False, **kwargs)

def SatResult(**kwargs):
    return Result(True, **kwargs)



from .backend_z3 import BackendZ3
from .backend_z3_parallel import BackendZ3Parallel
from .backend_concrete import BackendConcrete
from .backend_vsa import BackendVSA
from ..backend import Backend, BackendError



import logging
import functools

from .decorators import expand_ifproxy
from ..backend_object import BackendObject

l = logging.getLogger('claripy.vsa.ifproxy')

def proxified(f):
    @functools.wraps(f)
    def expander(self, *args):
        '''
        :param args: All arguments
        :return:
        '''
        op_name = f.__name__
        if_exprs = [ self.trueexpr, self.falseexpr ]
        ret = [ ]

        if isinstance(args[0], IfProxy):
            # Are the conditions opposite?
            other = args[0]
            cond1 = self.condition
            cond2 = other.condition

            if cond1.op in inverse_operations and \
                    inverse_operations[cond1.op] == cond2.op and \
                    all([ BoolResult.is_true(a1.identical(a2)) for a1, a2 in zip(cond1.args, cond2.args)]):
                other_args = [ other.falseexpr, other.trueexpr ]
            else:
                # FIXME: We are assuming the conditions are the same with self.condition...
                other_args = [ other.trueexpr, other.falseexpr ]
        else:
            other_args = [ args[0], args[0]  ]

        for i, arg in enumerate(if_exprs):
            if len(args) == 0:
                obj = NotImplemented

                if hasattr(arg, op_name):
                    op = getattr(arg, op_name)
                    obj = op()

                if obj is NotImplemented:
                    l.error('%s %s doesn\'t apply in IfProxy.expander()', self, op_name)
                    raise BackendError('Unable to apply operation on provided arguments.')

            else:
                obj = NotImplemented
                o = other_args[i]

                # first, try the operation with the first guy
                if hasattr(arg, op_name):
                    op = getattr(arg, op_name)
                    obj = op(o)
                # now try the reverse operation with the second guy
                if obj is NotImplemented and hasattr(o, opposites[op_name]):
                    op = getattr(o, opposites[op_name])
                    obj = op(arg)

                if obj is NotImplemented:
                    l.error("%s neither %s nor %s apply in IfProxy.expander()", self, op_name, opposites[op_name])
                    raise BackendError("unable to apply operation on provided arguments.")

                ret.append(obj)

        return IfProxy(self.condition, ret[0], ret[1])

    return expander

class IfProxy(BackendObject):
    def __init__(self, cond, true_expr, false_expr):
        self._cond = cond
        self._true = true_expr
        self._false = false_expr

    @staticmethod
    def unwrap(ifproxy, side=True):
        '''

        :param ifproxy:
        :param side: Decides which expr you want
        :return: A tuple of condition and expr
        '''

        # FIXME: find a better way to deal with cross-or'ed conditions!
        if isinstance(ifproxy, IfProxy):
            if side:
                return ifproxy.condition, IfProxy.unwrap(ifproxy.trueexpr, side)[1]
            else:
                return ifproxy.condition, IfProxy.unwrap(ifproxy.falseexpr, side)[1]
        else:
            return None, ifproxy

    @property
    def condition(self):
        return self._cond

    @property
    def trueexpr(self):
        return self._true

    @property
    def falseexpr(self):
        return self._false

    def __len__(self):
        return len(self._true)

    def __repr__(self):
        return 'IfProxy(%s, %s, %s)' % (self._cond, self._true, self._false)

    def identical(self, o):
        return self._true.identical(o._true) and self._false.identical(o._false) and self._cond.identical(o._cond)

    @proxified
    def __eq__(self, other): pass

    @proxified
    def __ne__(self, other): pass

    @proxified
    def __neg__(self): pass

    @proxified
    def __add__(self, other): pass

    @proxified
    def __radd__(self, other): pass

    @proxified
    def __sub__(self, other): pass

    @proxified
    def __rsub__(self, other): pass

    def __invert__(self):
        return IfProxy(self.condition, self.trueexpr.__invert__(), self.falseexpr.__invert__())

    @proxified
    def __or__(self, other): pass

    @proxified
    def __ror__(self, other): pass

    @proxified
    def __xor__(self, other): pass

    @proxified
    def __rxor__(self, other): pass

    @proxified
    def __and__(self, other): pass

    @proxified
    def __rand__(self, other): pass

from ..errors import BackendError
from ..operations import opposites, inverse_operations
from .bool_result import BoolResult



#!/usr/bin/env python

# pylint: disable=F0401,W0401,W0603,

import os
import sys
import socket
import logging
l = logging.getLogger("claripy")
l.addHandler(logging.NullHandler())

_all_backends = [ ]
_eager_backends = [ ]
_model_backends = [ ]

from .errors import *
from . import operations
from . import ops as _all_operations

from . import backends as _backends
backend_vsa = _backends.BackendVSA()

if not os.environ.get('WORKER', False) and os.environ.get('REMOTE', False):
    try:
        backend_z3 = _backends.backendremote.BackendRemote()
    except socket.error:
        raise ImportError("can't connect to backend")
else:
    backend_z3 = _backends.BackendZ3()

backend_concrete = _backends.BackendConcrete()

_eager_backends[:] = [ backend_concrete ]
_model_backends[:] = [ backend_concrete, backend_vsa ]
_all_backends[:] = [ backend_concrete, backend_vsa, backend_z3 ]
_backends = { 'BackendVSA': backend_vsa, 'BackendZ3': backend_z3, 'BackendConcrete': backend_concrete }

#
# connect to ANA
#

import ana
if os.environ.get('REMOTE', False):
    ana.set_dl(mongo_args=())

#
# Some other misguided setup
#

_recurse = 15000
l.warning("Claripy is setting the recursion limit to %d. If Python segfaults, I am sorry.", _recurse)
sys.setrecursionlimit(_recurse)

#
# Below are some exposed interfaces for general use.
#

def downsize():
    backend_vsa.downsize()
    backend_concrete.downsize()
    backend_z3.downsize()

#
# solvers
#

from .frontend import Frontend as _Frontend
from .frontends import LightFrontend, FullFrontend, CompositeFrontend
def Solver():
    return FullFrontend(backend_z3)
from .result import Result

#
# backend objects
#

from . import bv
from . import fp
from . import vsa

#
# Operations
#

from .ast.base import *
from .ast.bv import *
from .ast.fp import *
from .ast.bool import *
from . import ast
ast._import()

#
# and some aliases
#

BVV = BitVecVal
BV = BitVec
VS = ValueSet
SI = StridedInterval
TSI = TopStridedInterval
ESI = EmptyStridedInterval



import functools

def expand_ifproxy(f):
    @functools.wraps(f)
    def expander(*args, **kwargs):
        '''
        For each IfProxy proxified argument, we expand it so that it is
        converted into two operands (true expr and false expr, respectively).
        After the operation, we rewrap the two sides together in a new
        IfProxy with the old cond.

        :param args: All arguments
        :return:
        '''

        # FIXME: Now we have a very bad assumption - if we see more than one IfProxy
        # instances as the two operands, we assume they must both be true or
        # false.
        any_ifproxy = any([isinstance(a, IfProxy) for a in args])
        if any_ifproxy:
            true_args = None
            false_args = None
            cond = None
            # build true_args and false_args
            for a in args:
                cond = a.condition if isinstance(a, IfProxy) and cond is None else cond
                this_true_arg = IfProxy.unwrap(a, True)[1] if isinstance(a, IfProxy) else a
                this_false_arg = IfProxy.unwrap(a, False)[1] if isinstance(a, IfProxy) else a
                true_args = (this_true_arg, ) if true_args is None else true_args + (this_true_arg, )
                false_args = (this_false_arg, ) if false_args is None else false_args + (this_false_arg, )

            trueexpr = f(*true_args, **kwargs)
            falseexpr = f(*false_args, **kwargs)

            return IfProxy(cond, trueexpr, falseexpr)

        else:
            return f(*args, **kwargs)

    return expander

def expr_op_expand_ifproxy(f):
    @functools.wraps(f)
    def expander(self, *args, **kwargs):
        '''
        For each IfProxy proxified argument, we expand it so that it is
        converted into two operands (true expr and false expr, respectively).
        After the operation, we rewrap the two sides together in a new
        IfProxy with the old cond.

        :param args: All arguments
        :return:
        '''

        # FIXME: Now we have a very bad assumption - if we see two IfProxy
        # instances as the two operands, we assume they must both be true or
        # false.
        any_ifproxy = any([isinstance(a, Base) and isinstance(a.model, IfProxy) for a in args])
        if any_ifproxy:
            true_args = None
            false_args = None
            cond = None
            # build true_args and false_args
            for a in args:
                ifproxy = a.model if isinstance(a, Base) and isinstance(a.model, IfProxy) else None
                cond = ifproxy.condition if ifproxy is not None and cond is None else cond
                this_true_arg = IfProxy.unwrap(ifproxy, True)[1] if ifproxy is not None else a.model
                this_false_arg = IfProxy.unwrap(ifproxy, False)[1] if ifproxy is not None else a.model
                true_args = (this_true_arg, ) if true_args is None else true_args + (this_true_arg, )
                false_args = (this_false_arg, ) if false_args is None else false_args + (this_false_arg, )

            trueexpr = f(self, *true_args, **kwargs)
            falseexpr = f(self, *false_args, **kwargs)

            return IfProxy(cond, trueexpr, falseexpr)

        else:
            return f(self, *args, **kwargs)

    return expander

from .ifproxy import IfProxy
from ..ast.base import Base



import logging
l = logging.getLogger('claripy.ast.bool')

from ..ast.base import Base

class Bool(Base):
    @staticmethod
    def _from_bool(like, val): #pylint:disable=unused-argument
        return BoolI(val)

    def is_true(self):
        '''
        Returns True if 'self' can be easily determined to be True.
        Otherwise, return False. Note that the AST *might* still be True (i.e.,
        if it were simplified via Z3), but it's hard to quickly tell that.
        '''
        return is_true(self)

    def is_false(self):
        '''
        Returns True if 'self' can be easily determined to be False.
        Otherwise, return False. Note that the AST *might* still be False (i.e.,
        if it were simplified via Z3), but it's hard to quickly tell that.
        '''
        return is_false(self)

    def _simplify_And(self):
        if any(a.is_false() for a in self.args):
            return BoolI(False)
        else:
            new_args = [ a.simplified for a in self.args if not a.is_true() ]
            if len(new_args) == 0:
                return BoolI(True)
            else:
                return Bool(self.op, new_args)

    def _simplify_Or(self):
        if any(a.is_true() for a in self.args):
            return BoolI(True)
        else:
            new_args = [ a.simplified for a in self.args if not a.is_false() ]
            if len(new_args) == 0:
                return BoolI(False)
            else:
                return Bool(self.op, new_args)

def BoolI(model, **kwargs):
    return Bool('I', (model,), **kwargs)

def BoolVal(val):
    return BoolI(val, variables=set(), symbolic=False, eager=True)

#
# some standard ASTs
#

true = BoolVal(True)
false = BoolVal(False)

#
# Bound operations
#

from .. import operations

Bool.__eq__ = operations.op('__eq__', (Bool, Bool), Bool)
Bool.__ne__ = operations.op('__ne__', (Bool, Bool), Bool)

#
# Unbound operations
#

def If(*args):
    # the coercion here is strange enough that we'll just implement it manually
    if len(args) != 3:
        raise ClaripyOperationError("invalid number of args passed to If")

    args = list(args)

    if isinstance(args[0], bool):
        args[0] = BoolI(args[0], variables=frozenset(), symbolic=False, eager=True)

    ty = None
    if isinstance(args[1], Base):
        ty = type(args[1])
    elif isinstance(args[2], Base):
        ty = type(args[2])
    else:
        raise ClaripyTypeError("true/false clause of If must have bearable types")

    if not isinstance(args[1], ty):
        if hasattr(ty, '_from_' + type(args[1]).__name__):
            convert = getattr(ty, '_from_' + type(args[1]).__name__)
            args[1] = convert(args[2], args[1])
        else:
            raise ClaripyTypeError("can't convert {} to {}".format(type(args[1]), ty))
    if not isinstance(args[2], ty):
        if hasattr(ty, '_from_' + type(args[2]).__name__):
            convert = getattr(ty, '_from_' + type(args[2]).__name__)
            args[2] = convert(args[1], args[2])
        else:
            raise ClaripyTypeError("can't convert {} to {}".format(type(args[2]), ty))

    if is_true(args[0]):
        return args[1].make_like(args[1].op, args[1].args,
                                 variables=(args[1].variables | args[0].variables),
                                 symbolic=args[1].symbolic)
    elif is_false(args[0]):
        return args[2].make_like(args[2].op, args[2].args,
                                 variables=(args[2].variables | args[0].variables),
                                 symbolic=args[2].symbolic)

    if issubclass(ty, Bits):
        return ty('If', tuple(args), length=args[1].length)
    else:
        return ty('If', tuple(args)).reduced

And = operations.op('And', Bool, Bool, bound=False)
Or = operations.op('Or', Bool, Bool, bound=False)
Not = operations.op('Not', (Bool,), Bool, bound=False)

def is_true(e):
    for b in _all_backends:
        try: return b.is_true(b.convert(e))
        except BackendError: pass

    l.debug("Unable to tell the truth-value of this expression")
    return False

def is_false(e):
    for b in _all_backends:
        try: return b.is_false(b.convert(e))
        except BackendError: pass

    l.debug("Unable to tell the truth-value of this expression")
    return False

def ite_dict(i, d, default):
    return ite_cases([ (i == c, v) for c,v in d.items() ], default)

def ite_cases(cases, default):
    sofar = default
    for c,v in reversed(cases):
        sofar = If(c, v, sofar)
    return sofar

def constraint_to_si(expr):
    '''
    Convert a constraint to SI if possible
    :param expr:
    :return:
    '''

    satisfiable = True
    replace_list = [ ]

    satisfiable, replace_list = _all_backends[1].constraint_to_si(expr)

    # Make sure the replace_list are all ast.bvs
    for i in xrange(len(replace_list)):
        ori, new = replace_list[i]
        if not isinstance(new, Base):
            new = BVI(new, variables={ new.name }, symbolic=False, length=new._bits, eager=False)
            replace_list[i] = (ori, new)

    return satisfiable, replace_list

from .. import _all_backends
from ..errors import ClaripyOperationError, ClaripyTypeError, BackendError
from .bits import Bits
from .bv import BVI



import logging
import functools

l = logging.getLogger("claripy.backends.backend_vsa")

from ..backend import Backend, BackendError
from ..vsa import expand_ifproxy, expr_op_expand_ifproxy

def arg_filter(f):
    @functools.wraps(f)
    def filter(*args): #pylint:disable=redefined-builtin
        if type(args[0]) in {int, long}: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported argument type %s' % type(args[0]))
        return f(*args)

    return filter

def normalize_arg_order(f):
    @functools.wraps(f)
    def normalizer(*args):
        if len(args) != 2:
            raise BackendError('Unsupported arguments number %d' % len(args))

        if type(args[0]) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            if type(args[1]) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
                raise BackendError('Unsupported arguments')
            args = [args[1], args[0]]

        return f(*args)

    return normalizer

def normalize_boolean_arg_types(f):
    def convert_bool(a):
        if isinstance(a, BoolResult):
            return a
        if a == True:
            return TrueResult()
        elif a == False:
            return FalseResult()
        else:
            raise BackendError('Unsupported type %s' % type(a))

    @functools.wraps(f)
    def normalizer(*args):
        new_args = [convert_bool(a) for a in args]

        return f(*new_args)

    return normalizer

def convert_bvv_args(f):
    def convert_bvv(a):
        if isinstance(a, BVV):
            return BackendVSA.CreateStridedInterval(to_conv=a)
        return a

    @functools.wraps(f)
    def converter(*args):
        new_args = [convert_bvv(a) for a in args]

        return f(*new_args)

    return converter

def normalize_reversed_arguments(f):
    @functools.wraps(f)
    def normalizer(self, *args, **kwargs):
        arg_reversed = []
        raw_args = []
        for i in xrange(len(args)):
            if isinstance(args[i], Base) and \
                            type(args[i].model) in { #pylint:disable=unidiomatic-typecheck
                                                    StridedInterval,
                                                    DiscreteStridedIntervalSet,
                                                    ValueSet
            }:
                if args[i].model.reversed:
                    arg_reversed.append(True)
                    raw_args.append(args[i].reversed)
                    continue

            # It's not reversed
            arg_reversed.append(False)
            raw_args.append(args[i])

        any_reversed_arg = any(arg_reversed)
        for i in xrange(len(raw_args)):
            raw_args[i] = self.convert(raw_args[i])

        ret = f(self, *raw_args, **kwargs)

        variables = set()
        for a in raw_args:
            if isinstance(a, Base):
                variables |= a.variables
            else:
                variables.add(a.name)

        # inner_i = I(args[0]._claripy, ret, variables=variables)
        if any_reversed_arg:
            return ret.reverse()
            #ret = A(args[0]._claripy, 'Reverse', (inner_i,), variables=variables, collapsible=False)

        return ret

    return normalizer

class BackendVSA(Backend):
    def __init__(self):
        Backend.__init__(self)
        # self._make_raw_ops(set(expression_operations) - set(expression_set_operations), op_module=BackendVSA)
        self._make_expr_ops(set(expression_set_operations), op_class=self)
        self._make_raw_ops(set(backend_operations_vsa_compliant), op_module=BackendVSA)

        self._op_raw['StridedInterval'] = BackendVSA.CreateStridedInterval
        self._op_raw['ValueSet'] = ValueSet.__init__
        self._op_raw['AbstractLocation'] = AbstractLocation.__init__
        self._op_raw['Reverse'] = BackendVSA.Reverse
        self._op_expr['If'] = self.If

    def _convert(self, a, result=None):
        if type(a) in { int, long, float, bool, str }: #pylint:disable=unidiomatic-typecheck
            return a
        if type(a) is BVV: #pylint:disable=unidiomatic-typecheck
            return BackendVSA.CreateStridedInterval(bits=a.bits, to_conv=a)
        if type(a) in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            return a
        if isinstance(a, BoolResult):
            return a
        if isinstance(a, IfProxy):
            return a

        raise BackendError("why is fish raising NotImplementedError INSTEAD OF THE ERROR THAT'S SUPPOSED TO BE RAISED IN THIS SITUATION? SERIOUSLY, JUST RAISE A BACKENDERROR AND EVERYONE WILL BE HAPPY, BUT NO, PEOPLE HAVE TO RAISE THEIR OWN ERRORS INSTEAD OF USING THE ERRORS THAT WERE ****DESIGNED**** FOR THIS SORT OF THING. WHY DO I BOTHER DESIGNING A GOOD ERROR HIERARCHY, ANYWAYS? WILL IT BE USED? NO! IT'LL BE ALL NotImplementedError('THIS') or Exception('THAT') AND EVERYTHING WILL MELT DOWN. UGH!")

    def _eval(self, expr, n, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, StridedInterval):
            return expr.eval(n)
        elif isinstance(expr, ValueSet):
            results = []

            while len(results) < n:
                results.extend(expr.eval(n - len(results)))

            return results
        elif isinstance(expr, BoolResult):
            return expr.value
        elif isinstance(expr, IfProxy):
            results = set(self._eval(expr.trueexpr, n, result=result))
            if len(results) < n:
                results |= set(self._eval(expr.falseexpr, n - len(results), result=result))
            return list(results)
        else:
            raise BackendError('Unsupported type %s' % type(expr))

    def _min(self, expr, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, IfProxy):
            v1 = self.min(expr.trueexpr)
            v2 = self.min(expr.falseexpr)

            return min(v1, v2)

        elif isinstance(expr, StridedInterval):
            if expr.is_top:
                # TODO: Return
                return StridedInterval.min_int(expr.bits)

            return expr.lower_bound
        else:
            raise BackendError('Unsupported expr type %s' % type(expr))

    def _max(self, expr, result=None, solver=None, extra_constraints=()):
        if isinstance(expr, IfProxy):
            v1 = self.max(expr.trueexpr)
            v2 = self.max(expr.falseexpr)

            return max(v1, v2)

        else:
            if isinstance(expr, StridedInterval):
                if expr.is_top:
                    # TODO:
                    return StridedInterval.max_int(expr.bits)

                return expr.upper_bound

            else:
                raise BackendError('Unsupported expr type %s' % type(expr))

    def _solution(self, obj, v, result=None, solver=None, extra_constraints=()):
        if isinstance(obj, IfProxy):
            ret = self._solution(obj.trueexpr, v, result=result) or \
                self._solution(obj.falseexpr, v, result=result)
            return ret

        if isinstance(obj, BoolResult):
            return v in obj.value

        if isinstance(obj, StridedInterval):
            return not obj.intersection(v).is_empty

        if isinstance(obj, ValueSet):
            for _, si in obj.items():
                if not si.intersection(v).is_empty:
                    return True
            return False

        raise NotImplementedError(type(obj).__name__)

    def _has_true(self, o):
        return BoolResult.has_true(o)

    def _has_false(self, o):
        return BoolResult.has_false(o)

    def _is_true(self, o):
        return BoolResult.is_true(o)

    def _is_false(self, o):
        return BoolResult.is_false(o)

    #
    # Dealing with constraints
    #

    reversed_operations = { }
    reversed_operations['__ne__'] = '__eq__'
    reversed_operations['__eq__'] = '__ne__'
    reversed_operations['ULT'] = 'UGE'
    reversed_operations['UGT'] = 'ULE'
    reversed_operations['ULE'] = 'UGT'
    reversed_operations['UGE'] = 'ULT'
    reversed_operations['SLT'] = 'SGE'
    reversed_operations['SLE'] = 'SGT'
    reversed_operations['SGT'] = 'SLE'
    reversed_operations['SGE'] = 'SLT'
    reversed_operations['__le__'] = '__gt__'
    reversed_operations['__lt__'] = '__ge__'
    reversed_operations['__ge__'] = '__lt__'
    reversed_operations['__gt__'] = '__le__'

    comparison_info = { }
    # Tuples look like (is_lt, is_eq, is_unsigned)
    comparison_info['SLT'] = (True, False, False)
    comparison_info['SLE'] = (True, True, False)
    comparison_info['SGT'] = (False, False, False)
    comparison_info['SGE'] = (False, True, False)
    comparison_info['ULT'] = (True, False, True)
    comparison_info['ULE'] = (True, True, True)
    comparison_info['UGT'] = (False, False, True)
    comparison_info['UGE'] = (False, True, True)
    comparison_info['__lt__'] = comparison_info['ULT']
    comparison_info['__le__'] = comparison_info['ULE']
    comparison_info['__gt__'] = comparison_info['UGT']
    comparison_info['__ge__'] = comparison_info['UGE']

    def cts_simplifier_ZeroExt(self, args, expr, condition):
        """
        :param args:
        :param expr:
        :return: (new ast, new condition)
        """
        cond_op, cond_arg = condition
        # Normalize cond_arg
        if type(cond_arg) in (int, long): #pylint:disable=unidiomatic-typecheck
            cond_arg = _all_operations.BVV(cond_arg, expr.size())

        extended_bits, arg = args

        if cond_arg.size() <= arg.size() or \
                _all_operations.is_true(cond_arg[ expr.size() - 1 : expr.size() - extended_bits ] == 0):
            # We can safely eliminate this layer of ZeroExt
            if cond_arg.size() < arg.size():
                larger_cond_arg = cond_arg.zero_extend(arg.size() - cond_arg.size())
                if not isinstance(larger_cond_arg.model, Base):
                    return self.cts_simplify(arg.op, arg.args, arg, (cond_op, larger_cond_arg))
            else:
                return self.cts_simplify(arg.op, arg.args, arg, (cond_op, cond_arg[ arg.size() - 1 : 0 ]))

        else:
            # TODO: We may also handle the '__eq__' and '__ne__' case
            #__import__('ipdb').set_trace()
            # We cannot handle this...
            return expr, condition

    def cts_simplifier_SignExt(self, args, expr, condition):
        """

        :param args:
        :param expr:
        :param condition:
        :return:
        """
        # TODO: Review the logic of this method
        cond_op, cond_arg = condition
        # Normalize them
        if type(cond_arg) in (int, long): #pylint:disable=unidiomatic-typecheck
            cond_arg = _all_operations.BVV(cond_arg, expr.size())

        extended_bits, arg = args

        if cond_arg.size() <= arg.size() or \
                _all_operations.is_true(cond_arg[expr.size() - 1: expr.size() - extended_bits] == 0):
            # We can safely eliminate this layer of SignExt
            if cond_arg.size() < arg.size():
                larger_cond_arg = cond_arg.zero_extend(arg.size() - cond_arg.size()).resolved()
                if not isinstance(larger_cond_arg, Base):
                    return self.cts_simplify(arg.op, arg.args, arg, (cond_op, larger_cond_arg))
            else:
                return self.cts_simplify(arg.op, arg.args, arg, (cond_op, cond_arg[arg.size() - 1: 0]))

        else:
            # TODO: We may also handle the '__eq__' and '__ne__' case
            # __import__('ipdb').set_trace()
            # We cannot handle this...
            return expr, condition

    def cts_simplifier_Extract(self, args, expr, condition):
        '''
        Convert Extract(a, b, If(...)) to If(..., Extract(a, b, ...), Extract(a, b, ...))

        :param args:
        :return:
        '''

        high, low, to_extract = args
        ast, cond = self.cts_simplify(to_extract.op, to_extract.args, to_extract, condition)

        # Create the new ifproxy
        if ast is None:
            # We cannot handle it
            return None, condition

        elif ast.op == 'If':
            new_ifproxy = _all_operations.If(
                                ast.args[0],
                                _all_operations.Extract(high, low, ast.args[1]),
                                _all_operations.Extract(high, low, ast.args[2])
                            )

        else:
            cond_op, cond_arg = cond
            if type(cond_arg.model) in (int, long): #pylint:disable=unidiomatic-typecheck
                cond_arg = _all_operations.BVV(cond_arg, ast.size())
            elif type(cond_arg.model) in (StridedInterval, DiscreteStridedIntervalSet, BVV): #pylint:disable=unidiomatic-typecheck
                if ast.size() > cond_arg.size():
                    # Make sure two operands have the same size
                    cond_arg = _all_operations.ZeroExt(ast.size() - cond_arg.size(), cond_arg)

            if cond_arg.size() - 1 < high + 1 or \
                    _all_operations.is_true(cond_arg[cond_arg.size() - 1 : high + 1] == 0):
                # The upper part doesn't matter
                # We can handle it
                return self.cts_simplify(ast.op, ast.args, ast, (cond_op, cond_arg))
            else:
                # We cannot further simplify it
                return expr, condition

        return new_ifproxy, condition

    def cts_simplifier_Concat(self, args, expr, condition):
        '''
        Convert Concat(a, If(...)) to If(..., Concat(a, ...), Concat(a, ...))

        :param args:
        :return:
        '''

        new_args = [ self.cts_simplify(ex.op, ex.args, ex, condition) for ex in args ]

        ifproxy_conds = set([ a.args[0] for a, new_cond in new_args if a.op == 'If' ])

        if len(ifproxy_conds) == 0:
            # Let's check if we can remove this layer of Concat
            cond = condition[1]
            if len(args) == 2:
                if cond.size() - 1 >= cond.size() - args[0].size():
                    if _all_operations.is_true(args[0] == cond[ cond.size() - 1 : cond.size() - args[0].size() ]):
                        # Yes! We can remove it!
                        # TODO: This is hackish...
                        new_cond = (condition[0], cond[ cond.size() - args[0].size() - 1 : 0])
                        return self.cts_simplify(args[1].op, args[1].args, args[1], new_cond)

                else:
                    # args[0].size() == 0? It must be a bug.
                    raise ClaripyBackendVSAError(
                        'args[0].size() == %d (args[0] is \'%s\'). Please report this bug.' % (args[0].size, str(args[0])))

            # Cannot simplify it anymore
            return expr, condition

        elif len(ifproxy_conds) > 1:
            # We have more than one condition. Cannot handle it for now!
            return None, condition

        else:
            concat_trueexpr = [ ]
            concat_falseexpr = [ ]

            all_new_conds = set([ new_cond for a, new_cond in new_args ])

            if len(all_new_conds) > 1:
                # New conditions are not consistent. Can't handle it.
                return expr, condition

            for a, new_cond in new_args:
                if a.op == "If":
                    concat_trueexpr.append(a.args[1])
                    concat_falseexpr.append(a.args[2])
                else:
                    concat_trueexpr.append(a)
                    concat_falseexpr.append(a)

            new_ifproxy = _all_operations.If(
                                list(ifproxy_conds)[0],
                                _all_operations.Concat(*concat_trueexpr),
                                _all_operations.Concat(*concat_falseexpr)
                            )

            return new_ifproxy, condition

    def cts_simplifier_I(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use
        return expr, condition

    def cts_simplifier_If(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use
        return expr, condition

    def cts_simplifier_Reverse(self, args, expr, condition): #pylint:disable=unused-argument
        # TODO: How should we deal with Reverse in a smart way?

        arg = args[0]

        return self.cts_simplify(arg.op, arg.args, arg, condition)

    def cts_simplifier_widen(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def cts_simplifier_intersection(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def cts_simplifier___or__(self, args, expr, condition):
        claripy = expr._claripy
        argl, argr = args
        if argl is argr or claripy.is_true(argl == argr):
            return self.cts_simplify(argl.op, argl.args, argl, condition)
        elif claripy.is_true(argl == 0):
            return self.cts_simplify(argr.op, argr.args, argr, condition)
        elif claripy.is_true(argr == 0):
            return self.cts_simplify(argl.op, argl.args, argl, condition)
        else:
            return expr, condition

    def cts_simplifier___and__(self, args, expr, condition):

        argl, argr = args
        if argl is argr:
            # Operands are the same one!
            # We can safely remove this layer of __and__
            return self.cts_simplify(argl.op, argl.args, argl, condition)

        elif argl.structurally_match(argr):
            # Operands are the same
            # Safely remove the __and__ operation
            return self.cts_simplify(argl.op, argl.args, argl, condition)

        else:
            # We cannot handle it
            return expr, condition

    def cts_simplifier___xor__(self, args, expr, condition):
        argl, argr = args

        if _all_operations.is_true(argl == 0):
            # :-)
            return self.cts_simplify(argr.op, argr.args, argr, condition)
        elif _all_operations.is_true(argr == 0):
            # :-)
            return self.cts_simplify(argl.op, argl.args, argl, condition)
        else:
            # :-(
            return expr, condition

    def cts_simplifier___add__(self, args, expr, condition):

        argl, argr = args
        if _all_operations.is_true(argr == 0):
            # This layer of __add__ can be removed
            return self.cts_simplify(argl.op, argl.args, argl, condition)
        elif _all_operations.is_true(argl == 0):
            # This layer of __add__ can be removed
            return self.cts_simplify(argr.op, argr.args, argr, condition)
        else:

            if isinstance(argl.model, BVV):
                new_cond = (condition[0], condition[1] - argl)
                return self.cts_simplify(argr.op, argr.args, argr, new_cond)

            elif isinstance(argr.model, BVV):
                new_cond = (condition[0], condition[1] - argr)
                return self.cts_simplify(argl.op, argl.args, argl, new_cond)

            else:
                return expr, condition

    def cts_simplifier___radd__(self, args, expr, condition):
        return self.cts_simplifier___add__((args[1], args[0]), expr, condition)

    def cts_simplifier___sub__(self, args, expr, condition):
        """

        :param args:
        :param expr:
        :param condition:
        :return:
        """

        argl, argr = args
        if _all_operations.is_true(argr == 0):
            return self.cts_simplify(argl.op, argl.args, argl, condition)
        elif _all_operations.is_true(argl == 0):
            return self.cts_simplify(argr.op, argr.args, argr, condition)
        else:
            #__import__('ipdb').set_trace()
            return expr, condition

    def cts_simplifier___rsub__(self, args, expr, condition):
        return self.cts_simplifier___sub__((args[1], args[0]), expr, condition)

    def cts_simplifier___rshift__(self, args, expr, condition):

        arg, offset = args
        if _all_operations.is_true(offset == 0):
            return self.cts_simplify(arg.op, arg.args, arg, condition)
        else:
            return expr, condition

    def cts_simplifier___lshift__(self, args, expr, condition):

        arg, offset = args
        if _all_operations.is_true(offset == 0):
            return self.cts_simplify(arg.op, arg.args, arg, condition)
        else:
            return expr, condition

    def cts_simplifier___invert__(self, args, expr, condition):

        arg = args[0]
        if arg.op == 'If':
            new_arg = _all_operations.If(args[0], args[1].__invert__(), args[2].__invert__())

            return self.cts_simplify(new_arg.op, new_arg.args, expr, condition)

        else:
            __import__('ipdb').set_trace()
            return expr, condition

    def cts_simplifier_union(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def cts_simplifier___mod__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def cts_simplifier___div__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        return expr, condition

    def cts_simplifier___eq__(self, args, expr, condition): #pylint:disable=unused-argument,no-self-use

        l.error('cts_simplifier___eq__() should not exist. This is just a workaround for VSA. Fish will fix the issue later.')

        return expr, condition

    def cts_handler_comparison(self, args, comp=None):
        """
        Handles all comparisons.
        :param args:
        :param comp:
        :return:
        """

        if comp in self.comparison_info:
            is_lt, is_equal, is_unsigned = self.comparison_info[comp]
        else:
            raise ClaripyBackendVSAError('Support for comparison %s is not implemented. Please report it.' % comp)

        lhs, rhs = args

        if not isinstance(lhs, Base):
            raise ClaripyBackendVSAError('Left-hand-side expression is not an AST object.')

        # Maybe the target variable is the rhs
        if isinstance(lhs.model, BVV):
            new_op = self.reversed_operations[comp]
            new_lhs, new_rhs = rhs, lhs
            return self.cts_handle(new_op, (new_lhs, new_rhs))

        if type(rhs) in (int, long) or type(rhs.model) is BVV: #pylint:disable=unidiomatic-typecheck
            # Convert it into an SI
            rhs = _all_operations.SI(to_conv=rhs)

        if not isinstance(rhs, Base):
            raise ClaripyBackendVSAError('Right-hand-side expression cannot be converted to an AST object.')

        if lhs.op == 'If':
            condition, trueexpr, falseexpr = lhs.args

            if is_unsigned:
                if is_lt:
                    if is_equal:
                        take_true = _all_operations.is_true(trueexpr.ULE(rhs))
                        take_false = _all_operations.is_true(falseexpr.ULE(rhs))
                    else:
                        take_true = _all_operations.is_true(falseexpr.ULT(rhs))
                        take_false = _all_operations.is_true(trueexpr.ULT(rhs))
                else:
                    if is_equal:
                        take_true = _all_operations.is_true(trueexpr.UGE(rhs))
                        take_false = _all_operations.is_true(falseexpr.UGE(rhs))
                    else:
                        take_true = _all_operations.is_true(trueexpr.UGT(rhs))
                        take_false = _all_operations.is_true(falseexpr.UGT(rhs))
            else:
                if is_lt:
                    if is_equal:
                        take_true = _all_operations.is_true(trueexpr <= rhs)
                        take_false = _all_operations.is_true(falseexpr <= rhs)
                    else:
                        take_true = _all_operations.is_true(trueexpr < rhs)
                        take_false = _all_operations.is_true(falseexpr < rhs)
                else:
                    if is_equal:
                        take_true = _all_operations.is_true(trueexpr >= rhs)
                        take_false = _all_operations.is_true(falseexpr >= rhs)
                    else:
                        take_true = _all_operations.is_true(trueexpr > rhs)
                        take_false = _all_operations.is_true(falseexpr > rhs)

            if take_true and take_false:
                # It's always satisfiable
                return True, [ ]
            elif take_true:
                return self.cts_handle(condition.op, condition.args)
            elif take_false:
                rev_op = self.reversed_operations[condition.op]
                return self.cts_handle(rev_op, condition.args)
            else:
                # Not satisfiable
                return False, [ ]

        elif isinstance(rhs.model, StridedInterval) and isinstance(lhs.model, StridedInterval):
            if isinstance(lhs.model, Base):
                # It cannot be computed by our backend...
                # We just give up for now
                return True, [ ]

            stride = lhs.model.stride

            if is_lt:
                # < or <=
                if is_unsigned: lb = 0
                else: lb = StridedInterval.min_int(rhs.length)

                ub = rhs.model.upper_bound
                if not is_equal:
                    # <
                    ub = ub - 1

            else:
                # > or >=
                if is_unsigned: ub = StridedInterval.max_int(rhs.length)
                else: ub = StridedInterval.max_int(rhs.model.bits - 1)

                lb = rhs.model.lower_bound
                if not is_equal:
                    # >
                    lb = lb + 1

            if stride == 0 and lb != ub:
                # Make sure the final StridedInterval is always meaningful. See issue #55.
                stride = 1

            si_replacement = _all_operations.SI(bits=rhs.length, stride=stride, lower_bound=lb, upper_bound=ub)
            return True, [(lhs, si_replacement)]

        else:
            return True, [ ]

    def cts_handler___lt__(self, args): return self.cts_handler_comparison(args, comp='__lt__')
    def cts_handler___le__(self, args): return self.cts_handler_comparison(args, comp='__le__')
    def cts_handler___gt__(self, args): return self.cts_handler_comparison(args, comp='__gt__')
    def cts_handler___ge__(self, args): return self.cts_handler_comparison(args, comp='__ge__')
    def cts_handler_ULT(self, args): return self.cts_handler_comparison(args, comp='ULT')
    def cts_handler_ULE(self, args): return self.cts_handler_comparison(args, comp='ULE')
    def cts_handler_UGT(self, args): return self.cts_handler_comparison(args, comp='UGT')
    def cts_handler_UGE(self, args): return self.cts_handler_comparison(args, comp='UGE')
    def cts_handler_SLT(self, args): return self.cts_handler_comparison(args, comp='SLT')
    def cts_handler_SLE(self, args): return self.cts_handler_comparison(args, comp='SLE')
    def cts_handler_SGT(self, args): return self.cts_handler_comparison(args, comp='SGT')
    def cts_handler_SGE(self, args): return self.cts_handler_comparison(args, comp='SGE')

    def cts_handler_I(self, args): #pylint:disable=no-self-use
        a = args[0]

        if a in (False, 0):
            return False, [ ]
        elif isinstance(a, BVV) and a.value == 0:
            return False, [ ]

        return True, [ ]

    def cts_handler_Not(self, args):
        """
        The argument should be False

        :param args:
        :return:
        """

        a = args[0]
        expr_op = a.op
        expr_args = a.args

        # Reverse the op
        expr_op = self.reversed_operations[expr_op]

        return self.cts_handle(expr_op, expr_args)

    def cts_handler_And(self, args):
        """

        :param args:
        :return:
        """

        sat = True
        lst = [ ]

        # Both sides must be true
        for arg in args:
            sat_, lst_ = self.cts_handle(arg.op, arg.args)

            sat &= sat_
            lst.extend(lst_)

        if not sat:
            lst = [ ]

        return sat, lst

    def cts_handler_Or(self, args):

        if len(args) == 1:
            return self.cts_handle(args[0].op, args[0].args)

        else:
            if len(args) > 0:
                args = [ self.cts_handle(a.op, a.args) for a in args ]
                if any([not _all_operations.is_false(a) for a in args]):
                    return True, [ ]

                else:
                    return False, [ ]
            return True, [ ]

    def cts_handler___ne__(self, args):
        return self.cts_handler_eq_ne(args, False)

    def cts_handler___eq__(self, args):
        return self.cts_handler_eq_ne(args, True)

    def cts_handler_eq_ne(self, args, is_eq):
        """

        :param args:
        :return: True or False, and a list of (original_si, constrained_si) tuples
        """

        lhs, rhs = args

        if not isinstance(lhs, Base):
            raise ClaripyBackendVSAError('Left-hand-side expression is not an AST object.')

        size = lhs.size()

        if type(rhs) in (int, long): #pylint:disable=unidiomatic-typecheck
            # Convert it into a BVV
            rhs = _all_operations.BVV(rhs, size)

        if not isinstance(rhs, Base):
            raise ClaripyBackendVSAError('Right-hand-side expression cannot be converted to an AST object.')

        # TODO: Make sure the rhs doesn't contain any IfProxy

        if lhs.op == 'If':
            condition, trueexpr, falseexpr = lhs.args

            if is_eq:
                # __eq__
                take_true = _all_operations.is_true(rhs == trueexpr)
                take_false = _all_operations.is_true(rhs == falseexpr)
            else:
                # __ne__
                take_true = _all_operations.is_true(rhs == falseexpr)
                take_false = _all_operations.is_true(rhs == trueexpr)

            if take_true and take_false:
                # It's always satisfiable
                return True, [ ]

            elif take_true:
                # We take the true side
                return self.cts_handle(condition.op, condition.args)

            elif take_false:
                # We take the false side

                # Reverse the operation first
                rev_op = self.reversed_operations[condition.op]

                return self.cts_handle(rev_op, condition.args)

            else:
                # Not satisfiable
                return False, [ ]
        elif isinstance(lhs.model, StridedInterval) or isinstance(lhs.model, BVV):
            if not isinstance(lhs.model, StridedInterval):
                try: lhs = _all_operations.SI(to_conv=lhs)
                except BackendError: return True, [ ] # We cannot convert it to a StridedInterval

            try: rhs = _all_operations.SI(to_conv=rhs)
            except BackendError: return True, [ ]

            if is_eq:
                return True, [ (lhs, rhs)]
            else:
                if lhs.model.upper_bound <= rhs.model.upper_bound:
                    r = _all_operations.SI(bits=rhs.size(),
                                    stride=lhs.model.stride,
                                    lower_bound=lhs.model.lower_bound,
                                    upper_bound=rhs.model.lower_bound - 1)

                    return True, [ (lhs, r) ]
                elif lhs.model.lower_bound >= rhs.model.lower_bound:
                    r = _all_operations.SI(bits=rhs.size(),
                                    stride=lhs.model.stride,
                                    lower_bound=rhs.model.lower_bound + 1,
                                    upper_bound=lhs.model.upper_bound)

                    return True, [ (lhs, r) ]
                else:
                    # We cannot handle it precisely
                    return True, [ ]
        else:
            # TODO: handle this
            #import ipdb; ipdb.set_trace()

            return True, [ ]

    def cts_simplify(self, op, args, expr, condition):
        new_expr, new_cond = getattr(self, "cts_simplifier_%s" % op)(args, expr, condition)

        return new_expr, new_cond

    def cts_handle(self, op, args):

        if len(args) == 2:
            lhs, rhs = args

            # Simplify left side
            lhs, new_cond = self.cts_simplify(lhs.op, lhs.args, lhs, (op, rhs))

            # Update args
            op, rhs = new_cond
            args = (lhs, rhs)

            sat, lst = getattr(self, "cts_handler_%s" % op)(args)

        else:
            sat, lst = getattr(self, "cts_handler_%s" % op)(args)

        return sat, lst

    def constraint_to_si(self, expr):
        """
        We take in constraints, and convert them into constrained strided-intervals.

        For example, expr =
        <A __ne__ (
                    <A Extract (0,
                                0,
                                <A Concat (<A BVV(0x0, 63)>,
                                            <A If (
                                                    <A ULE (<A SI_1208<64>0x1[0x0, 0x100]>, <A BVV(0x27, 64)>)>,
                                                    <A BVV(0x1, 1)>,
                                                    <A BVV(0x0, 1)>
                                            )>
                                )>
                    )>,
                    0
        )>
        The result would be
        [ ( SI_1208<64>0x1[0x0, 0x100], SI_XXXX<64>0x1[0x0, 0x27] ) ]

        As we can only deal with bits, we will convert all integers into BVV during the solving and conversion process.

        :param expr: The constraint
        :return: whether the expr is satisfiable (boolean), and a list of tuples in form of
                (original_si, constrained_si).
        """

        try:
            sat, lst = self.cts_handle(expr.op, expr.args)

            return sat, lst

        except ClaripyVSASimplifierError as ex:
            l.error('VSASimplifiers raised an exception %s. Please report it.' % str(ex), exc_info=True)

            # return the dummy result
            return True, [ ]

    #
    # Backend Operations
    #

    def _identical(self, a, b, result=None):
        if type(a) != type(b):
            return False
        return a.identical(b)

    def _unique(self, obj, result=None): #pylint:disable=unused-argument,no-self-use
        if isinstance(obj, StridedInterval):
            return obj.unique
        elif isinstance(obj, ValueSet):
            return obj.unique
        elif isinstance(obj, IfProxy):
            return False
        else:
            raise BackendError('Not supported type of operand %s' % type(obj))

    def name(self, a, result=None):
        if isinstance(a, StridedInterval):
            return a.name

        else:
            return None

    @staticmethod
    @expand_ifproxy
    @normalize_boolean_arg_types
    def And(a, b):
        return a & b

    @staticmethod
    @expand_ifproxy
    @normalize_boolean_arg_types
    def Not(a):
        return ~a

    @staticmethod
    @normalize_arg_order
    def ULT(a, b):
        return a.ULT(b)

    @staticmethod
    @normalize_arg_order
    def ULE(a, b):
        return a.ULE(b)

    @staticmethod
    @normalize_arg_order
    def UGT(a, b):
        return a.UGT(b)

    @staticmethod
    @normalize_arg_order
    def UGE(a, b):
        return a.UGE(b)

    @staticmethod
    @normalize_arg_order
    def SLT(a, b):
        return a.SLT(b)

    @staticmethod
    @normalize_arg_order
    def SLE(a, b):
        return a.SLE(b)

    @staticmethod
    @normalize_arg_order
    def SGT(a, b):
        return a.SGT(b)

    @staticmethod
    @normalize_arg_order
    def SGE(a, b):
        return a.SGE(b)

    def If(self, cond, true_expr, false_expr, result=None): #pylint:disable=unused-argument
        exprs = []
        cond_model = self.convert(cond)
        if self.has_true(cond_model):
            exprs.append(true_expr)
        if self.has_false(cond_model):
            exprs.append(false_expr)

        if len(exprs) == 1:
            expr = self.convert(exprs[0])
        else:
            # TODO: How to handle it?
            expr = IfProxy(cond, self.convert(exprs[0]), self.convert(exprs[1]))

        return expr

    # TODO: Implement other operations!

    @staticmethod
    @expand_ifproxy
    def LShR(expr, shift_amount):
        return expr >> shift_amount

    @staticmethod
    @expand_ifproxy
    def Concat(*args):
        ret = None
        for expr in args:
            if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet, BVV }: #pylint:disable=unidiomatic-typecheck
                raise BackendError('Unsupported expr type %s' % type(expr))

            if type(expr) is BVV: #pylint:disable=unidiomatic-typecheck
                expr = BackendVSA.CreateStridedInterval(bits=expr.bits, to_conv=expr)

            ret = ret.concat(expr) if ret is not None else expr

        return ret

    @arg_filter
    def _size(self, arg, result=None):
        if type(arg) in { StridedInterval, DiscreteStridedIntervalSet, ValueSet, IfProxy }: #pylint:disable=unidiomatic-typecheck
            return len(arg)
        else:
            return arg.size()

    @staticmethod
    @expand_ifproxy
    def Extract(*args):
        low_bit = args[1]
        high_bit = args[0]
        expr = args[2]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet, ValueSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        ret = expr.extract(high_bit, low_bit)

        return ret

    @staticmethod
    @expand_ifproxy
    @convert_bvv_args
    def SignExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        return expr.sign_extend(new_bits + expr.bits)

    @staticmethod
    @expand_ifproxy
    @convert_bvv_args
    def ZeroExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in { StridedInterval, DiscreteStridedIntervalSet }: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(expr))

        return expr.zero_extend(new_bits + expr.bits)

    @staticmethod
    @expand_ifproxy
    def Reverse(arg):
        if type(arg) not in {StridedInterval, DiscreteStridedIntervalSet, ValueSet}: #pylint:disable=unidiomatic-typecheck
            raise BackendError('Unsupported expr type %s' % type(arg))

        return arg.reverse()

    @expr_op_expand_ifproxy
    @normalize_reversed_arguments
    def union(self, *args, **kwargs): #pylint:disable=unused-argument,no-self-use
        if len(args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.union().' % len(args))

        ret = args[0].union(args[1])

        if ret is NotImplemented:
            ret = args[1].union(args[0])

        return ret

    @expr_op_expand_ifproxy
    @normalize_reversed_arguments
    def intersection(self, *args, **kwargs): #pylint:disable=unused-argument,no-self-use
        if len(args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.intersection().' % len(args))

        ret = None

        for arg in args:
            if ret is None:
                ret = arg
            else:
                ret = ret.intersection(arg)

        return ret

    @expr_op_expand_ifproxy
    @normalize_reversed_arguments
    def widen(self, *args, **kwargs): #pylint:disable=unused-argument,no-self-use
        if len(args) != 2:
            raise BackendError('Incorrect number of arguments (%d) passed to BackendVSA.widen().' % len(args))

        ret = args[0].widen(args[1])
        if ret is NotImplemented:
            ret = args[1].widen(args[0])

        return ret

    @staticmethod
    def CreateTopStridedInterval(bits, name=None, uninitialized=False): #pylint:disable=unused-argument,no-self-use
        return StridedInterval.top(bits, name, uninitialized=uninitialized)

from ..bv import BVV
from ..ast.base import Base
from ..operations import backend_operations_vsa_compliant, expression_set_operations
from ..vsa import StridedInterval, CreateStridedInterval, DiscreteStridedIntervalSet, ValueSet, AbstractLocation, BoolResult, TrueResult, FalseResult
from ..vsa import IfProxy
from ..errors import ClaripyBackendVSAError, ClaripyVSASimplifierError

from .. import _all_operations

BackendVSA.CreateStridedInterval = staticmethod(CreateStridedInterval)
